import os
import re
import sys

from . import MiscUtils
from .MiscUtils import commands_get_output, commands_get_status_output_with_env, commands_get_output_with_env
from . import PLogger
from . import Client

# error code
EC_Config    = 10


# get CMT projects
def getCmtProjects(dir='.'):
    # cmt or cmake
    if not useCMake():
        # keep current dir
        curdir = os.getcwd()
        # change dir
        os.chdir(dir)
        # get projects
        out = commands_get_output_with_env('cmt show projects')
        lines = out.split('\n')
        # remove CMT warnings
        tupLines = tuple(lines)
        lines = []
        for line in tupLines:
            if 'CMTUSERCONTEXT' in line:
                continue
            if not line.startswith('#'):
                lines.append(line)
        # back to the current dir
        os.chdir(curdir)
        # return
        return lines,out         
    else:
        lines = []
        # use current dir as test area
        tmpStr = '(in {0})'.format(os.getcwd())
        lines.append(tmpStr)
        # AtlasProject
        if not 'AtlasProject' in os.environ:
            return [],'AtlasProject is not defined in runtime environment'
        tmpStr = '{0} (in {0}/{1})'.format(os.environ['AtlasProject'],os.environ['AtlasVersion'])
        lines.append(tmpStr)
        if 'AtlasOffline_VERSION' in os.environ:
            tmpStr = '{0} (in {0}/{1})'.format('AtlasOffline',os.environ['AtlasOffline_VERSION'])
            lines.append(tmpStr)
        prodVerStr = '{0}_VERSION'.format(os.environ['AtlasProject'])
        if prodVerStr in os.environ:
            tmpStr = '{0} (in {0}/{1})'.format(os.environ['AtlasProject'],os.environ[prodVerStr])
            lines.append(tmpStr)
        return lines,''
    

# check if ath release
def isAthRelease(cacheVer):
    try:
        if 'AthAnalysis' in cacheVer or re.search('Ath[a-zA-Z]+Base',cacheVer) is not None:
            return True
    except Exception:
        pass
    return False

    
# get Athena version
def getAthenaVer(verbose=True):
    # get logger
    tmpLog = PLogger.getPandaLogger()            
    # get project parameters
    lines,out = getCmtProjects()
    if len(lines)<2:
        # make a tmp dir to execute cmt
        tmpDir = 'cmttmp.%s' % MiscUtils.wrappedUuidGen()
        os.mkdir(tmpDir)
        # try cmt under a subdir since it doesn't work in top dir
        lines,tmpOut = getCmtProjects(tmpDir)
        # delete the tmp dir
        commands_get_output('rm -rf %s' % tmpDir)
        if len(lines)<2:
            if verbose:
                print(out)
                tmpLog.error("cmt gave wrong info")
            return False,{}
    # private work area
    res = re.search('\(in ([^\)]+)\)',lines[0])
    if res==None:
        print(lines[0])
        tmpLog.error("no TestArea. could not get path to private work area")
        return False,{}
    workArea = os.path.realpath(res.group(1))
    # get Athena version and group area
    athenaVer = ''
    groupArea = ''
    cacheVer  = ''
    nightVer  = ''
    cmtConfig = ''
    for line in lines[1:]:
        res = re.search('\(in ([^\)]+)\)',line)
        if res is not None:
            items = line.split()
            if items[0].startswith('Athena') or items[0].startswith('Analysis') or \
                    items[0] in ['AthDerivations','AnalysisBase','AthSimulation','AthDerivation', 'AthAnalysis',
                                 'AthGeneration', 'AtlasStats']:
                isGitBase = True
            else:
                isGitBase = False
            # base release
            if items[0] in ('dist','AtlasRelease','AtlasOffline','AtlasAnalysis','AtlasTrigger',
                            'AtlasReconstruction') or isGitBase:
                # Atlas release
                if 'AtlasBuildStamp' in os.environ and ('AtlasReleaseType' not in os.environ or
                                                            os.environ['AtlasReleaseType'] != 'stable'):
                    athenaVer = os.environ['AtlasBuildStamp']
                    useBuildStamp = True
                else:
                    athenaVer = os.path.basename(res.group(1))
                    useBuildStamp = False
                # nightly
                if athenaVer.startswith('rel') or useBuildStamp or isGitBase:
                    # extract base release
                    if not useCMake():
                        tmpMatch = re.search('/([^/]+)(/rel_\d+)*/Atlas[^/]+/rel_\d+',line)
                        if tmpMatch is None:
                            tmpLog.error("unsupported nightly %s" % line)
                            return False,{}
                        # set athenaVer and cacheVer
                        cacheVer  = '-AtlasOffline_%s' % athenaVer
                        athenaVer = tmpMatch.group(1)
                    else:
                        if athenaVer.startswith('rel'):
                            tmpLog.error("Nightlies with AFS setup are unsupported on the grid. Setup with CVMFS")
                            return False,{}
                        if isGitBase:
                            cacheVer  = '-{0}_{1}'.format(items[0],athenaVer)
                        else:                            
                            cacheVer  = '-AtlasOffline_%s' % athenaVer
                        if useBuildStamp:
                            athenaVer = os.environ['AtlasBuildBranch']
                break
            # cache or analysis projects
            elif items[0] in ['AtlasProduction','AtlasPoint1','AtlasTier0','AtlasP1HLT',
                              'AtlasDerivation','TrigMC'] or \
                              isAthRelease(items[0]) or \
                              items[1].count('.') >= 4:  
                # cache is used
                if cacheVer != '':
                    continue
                # production cache
                cacheTag = os.path.basename(res.group(1))
                if items[0] == 'AtlasProduction' and cacheTag.startswith('rel'):
                    # nightlies for cache
                    tmpMatch = re.search('/([^/]+)(/rel_\d+)*/Atlas[^/]+/rel_\d+',line)
                    if tmpMatch is None:
                        tmpLog.error("unsupported nightly %s" % line)
                        return False,{}
                    cacheVer  = '-AtlasOffline_%s' % cacheTag
                    athenaVer = tmpMatch.group(1)
                    break
                elif items[0] == 'TrigMC' and cacheTag.startswith('rel'):
                    # nightlies for cache
                    tmpMatch = re.search('/([^/]+)(/rel_\d+)*/[^/]+/rel_\d+',line)
                    if tmpMatch is None:
                        tmpLog.error("unsupported nightly %s" % line)
                        return False,{}
                    cacheVer  = '-%s_%s' % (items[0],cacheTag)
                    athenaVer = tmpMatch.group(1)
                    break
                elif isAthRelease(items[0]):
                    cacheVer  = '-%s_%s' % (items[0],cacheTag)
                else:
                    # doesn't use when it is a base release since it is not installed in EGEE
                    if re.search('^\d+\.\d+\.\d+$',cacheTag) is None:
                        cacheVer = '-%s_%s' % (items[0],cacheTag)
                # no more check for AthAnalysis
                if isAthRelease(items[0]):
                    break
            else:
                # group area
                groupArea = os.path.realpath(res.group(1))
    # cmtconfig
    if 'CMTCONFIG' in os.environ:
        cmtConfig = os.environ['CMTCONFIG']
    # last resort
    if athenaVer == '':
        if 'AtlasProject' in os.environ and 'AtlasBuildBranch' in os.environ:
            prodVerStr = '{0}_VERSION'.format(os.environ['AtlasProject'])
            if prodVerStr in os.environ:
                athenaVer = os.environ['AtlasBuildBranch']
                cacheVer = '-{0}_{1}'.format(os.environ['AtlasProject'],
                                             os.environ[prodVerStr])
                groupArea = ''
    # pack return values
    retVal = {
        'workArea' : workArea,
        'athenaVer': athenaVer,
        'groupArea': groupArea,
        'cacheVer' : cacheVer,
        'nightVer' : nightVer,
        'cmtConfig': cmtConfig,
           }
    # check error
    if athenaVer == '' and not isAthRelease(cacheVer):
        tmpStr = ''
        for line in lines:
            tmpStr += (line+'\n')
        tmpLog.info('cmt showed\n'+tmpStr)
        tmpLog.error("could not get Athena version. perhaps your requirements file doesn't have ATLAS_TEST_AREA")
        return False,retVal
    # return
    return True,retVal


# wrapper for attribute access
class ConfigAttr(dict):
    # override __getattribute__ for dot access
    def __getattribute__(self,name):
        if name in dict.__dict__.keys():
            return dict.__getattribute__(self,name)
        if name.startswith('__'):
            return dict.__getattribute__(self,name)
        if name in dict.keys(self):
            return dict.__getitem__(self,name)
        return False

    def __setattr__(self,name,value):
        if name in dict.__dict__.keys():
            dict.__setattr__(self,name,value)
        else:
            dict.__setitem__(self,name,value)
                        

# extract run configuration
def extractRunConfig(jobO,supStream,shipinput,trf,verbose=False,useAMI=False,
                     inDS='',tmpDir='.'):
    # get logger
    tmpLog = PLogger.getPandaLogger()
    outputConfig = ConfigAttr()
    inputConfig  = ConfigAttr()
    otherConfig  = ConfigAttr()
    statsCode = True
    if trf:
        pass
    else:
        # use AMI
        amiJobO = ''
        if useAMI:
            amiJobO = getJobOtoUseAmiForAutoConf(inDS,tmpDir)
        baseName = os.environ['PANDA_SYS'] + "/etc/panda/share"
        if ' - ' in jobO:
            jobO = re.sub(' - ', ' %s/ConfigExtractor.py - ' % baseName, jobO)
        else:
            jobO = jobO + ' %s/ConfigExtractor.py ' % baseName
        com = 'athena.py %s %s/FakeAppMgr.py %s' % \
              (amiJobO,baseName,jobO)
        if verbose:
            tmpLog.debug(com)
        # run ConfigExtractor for normal jobO
        out = commands_get_output_with_env(com)
        failExtractor = True
        outputConfig['alloutputs'] = []
        skipOutName = False
        for line in out.split('\n'):
            match = re.findall('^ConfigExtractor > (.+)',line)
            if len(match):
                # suppress some streams
                if match[0].startswith("Output="):
                    tmpSt0 = "NoneNoneNone"
                    tmpSt1 = "NoneNoneNone"
                    tmpSt2 = "NoneNoneNone"
                    try:
                        tmpSt0 = match[0].replace('=',' ').split()[1].upper()
                    except Exception:
                        pass
                    try:
                        tmpSt1 = match[0].replace('=',' ').split()[-1].upper()
                    except Exception:
                        pass
                    try:
                        tmpSt2 = match[0].replace('=',' ').split()[2].upper()
                    except Exception:
                        pass
                    toBeSuppressed = False
                    # normal check
                    if tmpSt0 in supStream or tmpSt1 in supStream or tmpSt2 in supStream:
                        toBeSuppressed = True
                    # wild card check
                    if not toBeSuppressed:
                        for tmpPatt in supStream:
                            if '*' in tmpPatt:
                                tmpPatt = '^' + tmpPatt.replace('*','.*')
                                tmpPatt = tmpPatt.upper()
                                if re.search(tmpPatt,tmpSt0) is not None or \
                                       re.search(tmpPatt,tmpSt1) is not None or \
                                       re.search(tmpPatt,tmpSt2) is not None:
                                    toBeSuppressed = True
                                    break
                    # suppressed            
                    if toBeSuppressed:            
                        tmpLog.info('%s is suppressed' % line)
                        # set skipOutName to ignore output filename in the next loop
                        skipOutName = True
                        continue
                failExtractor = False
                # AIDA HIST
                if match[0].startswith('Output=HIST'):
                    outputConfig['outHist'] = True
                # AIDA NTuple
                if match[0].startswith('Output=NTUPLE'):
                    if 'outNtuple' not in outputConfig:
                        outputConfig['outNtuple'] = []
                    tmpItems = match[0].split()
                    outputConfig['outNtuple'].append(tmpItems[1])
                # RDO
                if match[0].startswith('Output=RDO'):
                    outputConfig['outRDO'] = match[0].split()[1]
                # ESD
                if match[0].startswith('Output=ESD'):
                    outputConfig['outESD'] = match[0].split()[1]
                # AOD
                if match[0].startswith('Output=AOD'):
                    outputConfig['outAOD'] = match[0].split()[1]
                # TAG output
                if match[0]=='Output=TAG':            
                    outputConfig['outTAG'] = True
                # TAGCOM
                if match[0].startswith('Output=TAGX'):
                    if 'outTAGX' not in outputConfig:
                        outputConfig['outTAGX'] = []
                    tmpItems = match[0].split()
                    outputConfig['outTAGX'].append(tuple(tmpItems[1:]))
                # AANT
                if match[0].startswith('Output=AANT'):
                    if 'outAANT' not in outputConfig:
                        outputConfig['outAANT'] = []
                    tmpItems = match[0].split()
                    outputConfig['outAANT'].append(tuple(tmpItems[1:]))
                # THIST
                if match[0].startswith('Output=THIST'):            
                    if 'outTHIST' not in outputConfig:
                        outputConfig['outTHIST'] = []
                    tmpItems = match[0].split()
                    if not tmpItems[1] in outputConfig['outTHIST']:
                        outputConfig['outTHIST'].append(tmpItems[1])
                # IROOT
                if match[0].startswith('Output=IROOT'):            
                    if 'outIROOT' not in outputConfig:
                        outputConfig['outIROOT'] = []
                    tmpItems = match[0].split()
                    outputConfig['outIROOT'].append(tmpItems[1])
                # Stream1
                if match[0].startswith('Output=STREAM1'):
                    outputConfig['outStream1'] = match[0].split()[1]
                # Stream2
                if match[0].startswith('Output=STREAM2'):                        
                    outputConfig['outStream2'] = match[0].split()[1]
                # ByteStream output
                if match[0]=='Output=BS':            
                    outputConfig['outBS'] = True
                # General Stream
                if match[0].startswith('Output=STREAMG'):            
                    tmpItems = match[0].split()
                    outputConfig['outStreamG'] = []
                    for tmpNames in tmpItems[1].split(','):
                        outputConfig['outStreamG'].append(tmpNames.split(':'))
                # Metadata
                if match[0].startswith('Output=META'):            
                    if 'outMeta' not in outputConfig:
                        outputConfig['outMeta'] = []
                    tmpItems = match[0].split()
                    outputConfig['outMeta'].append(tuple(tmpItems[1:]))
                # UserDataSvc
                if match[0].startswith('Output=USERDATA'):            
                    if 'outUserData' not in outputConfig:
                        outputConfig['outUserData'] = []
                    tmpItems = match[0].split()
                    outputConfig['outUserData'].append(tmpItems[-1])
                # MultipleStream
                if match[0].startswith('Output=MS'):
                    if 'outMS' not in outputConfig:
                        outputConfig['outMS'] = []
                    tmpItems = match[0].split()            
                    outputConfig['outMS'].append(tuple(tmpItems[1:]))
                # No input
                if match[0]=='No Input':
                    inputConfig['noInput'] = True
                # ByteStream input
                if match[0]=='Input=BS':                        
                    inputConfig['inBS'] = True
                # selected ByteStream
                if match[0].startswith('Output=SelBS'):            
                    tmpItems = match[0].split()
                    inputConfig['outSelBS'] = tmpItems[1]
                # TAG input
                if match[0]=='Input=COLL':                        
                    inputConfig['inColl'] = True
                # POOL references
                if match[0].startswith('Input=COLLREF'):
                    tmpRef = match[0].split()[-1]
                    if tmpRef == 'Input=COLLREF':
                        # use default token when ref is empty
                        tmpRef = 'Token'
                    elif tmpRef != 'Token' and (not tmpRef.endswith('_ref')):
                        # append _ref
                        tmpRef += '_ref'
                    inputConfig['collRefName'] = tmpRef
                # TAG Query
                if match[0].startswith('Input=COLLQUERY'):
                    tmpQuery = re.sub('Input=COLLQUERY','',match[0])
                    tmpQuery = tmpQuery.strip()
                    inputConfig['tagQuery'] = tmpQuery
                # Minimum bias
                if match[0]=='Input=MINBIAS':
                    inputConfig['inMinBias'] = True
                # Cavern input
                if match[0]=='Input=CAVERN':
                    inputConfig['inCavern'] = True
                # Beam halo
                if match[0]=='Input=BEAMHALO':
                    inputConfig['inBeamHalo'] = True
                # Beam gas
                if match[0]=='Input=BEAMGAS':
                    inputConfig['inBeamGas'] = True
                # Back navigation
                if match[0]=='BackNavigation=ON':                        
                    inputConfig['backNavi'] = True
                # Random stream
                if match[0].startswith('RndmStream'):
                    if 'rndmStream' not in otherConfig:
                        otherConfig['rndmStream'] = []
                    tmpItems = match[0].split()
                    otherConfig['rndmStream'].append(tmpItems[1])
                # Generator file
                if match[0].startswith('RndmGenFile'):
                    if 'rndmGenFile' not in otherConfig:
                        otherConfig['rndmGenFile'] = []
                    tmpItems = match[0].split()
                    otherConfig['rndmGenFile'].append(tmpItems[-1])
                # G4 Random seeds
                if match[0].startswith('G4RandomSeeds'):
                    otherConfig['G4RandomSeeds'] = True
                # input files for direct input
                if match[0].startswith('InputFiles'):
                    if shipinput:
                        tmpItems = match[0].split()
                        otherConfig['inputFiles'] = tmpItems[1:]
                    else:
                        continue
                # condition file
                if match[0].startswith('CondInput'):
                    if 'condInput' not in otherConfig:
                        otherConfig['condInput'] = []
                    tmpItems = match[0].split()
                    otherConfig['condInput'].append(tmpItems[-1])
                # collect all outputs
                if match[0].startswith(' Name:'):
                    # skipped output
                    if skipOutName:
                        skipOutName = False
                        continue
                    outputConfig['alloutputs'].append(match[0].split()[-1])
                    continue
                tmpLog.info(line)
                skipOutName = False
        # extractor failed
        if failExtractor:
            print(out)
            tmpLog.error("Could not parse jobOptions")
            statsCode = False
    # return
    retConfig = ConfigAttr()
    retConfig['input']  = inputConfig
    retConfig['other']  = otherConfig
    retConfig['output'] = outputConfig
    return statsCode,retConfig


# extPoolRefs for old releases which don't contain CollectionTools
athenaStuff = []

# jobO files with full path names
fullPathJobOs = {}


# convert fullPathJobOs to str
def convFullPathJobOsToStr():
    tmpStr = ''
    for fullJobO in fullPathJobOs:
        localName = fullPathJobOs[fullJobO]
        tmpStr += '%s:%s,' % (fullJobO,localName)
    tmpStr = tmpStr[:-1]
    return tmpStr


# convert str to fullPathJobOs
def convStrToFullPathJobOs(tmpStr):
    retMap = {}
    for tmpItem in tmpStr.split(','):
        fullJobO,localName = tmpItem.split(':')
        retMap[fullJobO] = localName
    return retMap

        
# copy some athena specific files and full-path jobOs
def copyAthenaStuff(currentDir):
    baseName = os.environ['PANDA_SYS'] + "/etc/panda/share"
    for tmpFile in athenaStuff:
        com = 'cp -p %s/%s %s' % (baseName,tmpFile,currentDir)
        commands_get_output(com)
    for fullJobO in fullPathJobOs:
        localName = fullPathJobOs[fullJobO]
        com = 'cp -p %s %s/%s' % (fullJobO,currentDir,localName)
        commands_get_output(com)


# delete some athena specific files and copied jobOs
def deleteAthenaStuff(currentDir):
    for tmpFile in athenaStuff:
        com = 'rm -f %s/%s' % (currentDir,tmpFile)
        commands_get_output(com)
    for tmpFile in fullPathJobOs.values():
        com = 'rm -f %s/%s' % (currentDir,tmpFile)
        commands_get_output(com)


# set extFile
extFile = []
def setExtFile(v_extFile):
    global extFile
    extFile = v_extFile


# set excludeFile
excludeFile = []
def setExcludeFile(strExcludeFile):
    # empty
    if strExcludeFile == '':
        strExcludeFile = 'jobReport.json,jobReport.txt,jobReportExtract.pickle'
    else:
        strExcludeFile += ',jobReport.json,jobReport.txt,jobReportExtract.pickle'
    # convert to list
    global excludeFile
    for tmpItem in strExcludeFile.split(','):
        tmpItem = tmpItem.strip()
        if tmpItem == '':
            continue
        # change . to \. for regexp
        tmpItem = tmpItem.replace('.','\.')        
        # change * to .* for regexp
        tmpItem = tmpItem.replace('*','.*')
        # append
        excludeFile.append(tmpItem)
    
                
# matching for extFiles
def matchExtFile(fileName):
    # check exclude files
    for tmpPatt in excludeFile:
        if re.search(tmpPatt,fileName) is not None:
            return False
    # gather files with special extensions
    for tmpExtention in ['.py','.dat','.C','.xml','Makefile',
                         '.cc','.cxx','.h','.hh','.sh','.cpp',
                         '.hpp']:
        if fileName.endswith(tmpExtention):
            return True
    # check filename
    baseName = fileName.split('/')[-1]
    for patt in extFile:
        if patt.find('*') == -1:
            # regular matching
            if patt == baseName:
                return True
            # patt may contain / for sub dir
            if patt != '' and re.search(patt,fileName) is not None:
                return True
        else:
            # use regex for *
            tmpPatt = patt.replace('*','.*')
            if re.search(tmpPatt,baseName) is not None:
                return True
            # patt may contain / for sub dir
            if patt != '' and re.search(tmpPatt,fileName) is not None:
                return True
    # not matched
    return False


# extended extra stream name
useExtendedExtStreamName = False

# use extended extra stream name
def enableExtendedExtStreamName():
    global useExtendedExtStreamName
    useExtendedExtStreamName = True
        
# get extended extra stream name
def getExtendedExtStreamName(sIndex,sName,enableExtension):
    tmpBaseExtName = 'EXT%s' % sIndex
    if not useExtendedExtStreamName or not enableExtension:
        return tmpBaseExtName
    # change * to X and add .tgz
    if sName.find('*') != -1:
        sName = sName.replace('*','XYZ')
        sName = '%s.tgz' % sName
    # use extended extra stream name
    tmpItems = sName.split('.')
    if len(tmpItems) > 0:
        tmpBaseExtName += '_%s' % tmpItems[0]
    return tmpBaseExtName
    

# special files to be treated carefully
specialFilesForAthena = ['dblookup.xml']

# archive source files
def archiveSourceFiles(workArea,runDir,currentDir,tmpDir,verbose,gluePackages=[],dereferenceSymLinks=False,
                       archiveName=''):
    # archive sources
    tmpLog = PLogger.getPandaLogger()
    tmpLog.info('archiving source files')

    #####################################################################
    # subroutines

    # scan InstallArea to get a list of local packages
    def getFileList(dir,files,forPackage,readLink=True):
        try:
            list = os.listdir(dir)
        except Exception:
            return
        for item in list:
            # skip if doc or .svn
            if item in ['doc','.svn','_CPack_Packages']:
                continue
            fullName=dir+'/'+item
            if os.path.isdir(fullName):
                # ignore symlinked dir just under InstallArea/include
                # they are created for g77
                if os.path.islink(fullName) and re.search('/InstallArea/include$',dir) is not None:
                    pass
                elif os.path.islink(fullName) and readLink and forPackage:
                    # resolve symlink
                    getFileList(os.readlink(fullName),files,forPackage,readLink)
                else:
                    getFileList(fullName,files,forPackage,readLink)
            else:
                if os.path.islink(fullName):
                    if readLink:
                        tmpLink = os.readlink(fullName)
                        # put base dir when relative path
                        if not tmpLink.startswith('/'):
                            tmpLink = dir+'/'+tmpLink
                            tmpLink = os.path.abspath(tmpLink)
                        appFileName = tmpLink
                    else:
                        appFileName = os.path.abspath(fullName)                        
                else:
                    appFileName = os.path.abspath(fullName)
                # remove redundant //
                appFilename = re.sub('//','/',appFileName)
                # append
                files.append(appFileName)

    # get package list
    def getPackages(_workArea,gluePackages=[]):
        # get logger
        tmpLog = PLogger.getPandaLogger()
        # special packages
        specialPackages = {'External/Lhapdf':'external/MCGenerators/lhapdf'}
        # get file list
        installFiles = []
        getFileList(_workArea+'/InstallArea',installFiles,True)
        # get list of packages
        cmt_config = os.environ['CMTCONFIG']
        _packages = []
        for iFile in installFiles:
            # ignore InstallArea stuff
            if re.search('/InstallArea/',iFile):
                continue
            # converted to real path
            file = os.path.realpath(iFile)
            # remove special characters
            sString=re.sub('[\+]','.',os.path.realpath(_workArea))
            # look for /share/ , /python/, /i686-slc3-gcc323-opt/, .h
            for target in ('share/','python/',cmt_config+'/','[^/]+\.h'):
                res = re.search(sString+'/(.+)/'+target,file)
                if res:
                    # append
                    pName = res.group(1)
                    if target in ['[^/]+\.h']:
                        # convert PackageDir/PackageName/PackageName to PackageDir/PackageName
                        pName = re.sub('/[^/]+$','',pName)
                    if pName not in _packages:
                        if os.path.isdir(_workArea+'/'+pName):
                            _packages.append(pName)
                    break
            # check special packages just in case
            for pName in specialPackages:
                pPath = specialPackages[pName]
                if pName not in _packages:
                    # look for path pattern
                    if re.search(pPath,file) is not None:
                        if os.path.isdir(_workArea+'/'+pName):
                            # check structured style
                            tmpDirList = os.listdir(_workArea+'/'+pName)
                            useSS = False
                            for tmpDir in tmpDirList:
                                if re.search('-\d+-\d+-\d+$',tmpDir) is not None:
                                    _packages.append(pName+'/'+tmpDir)
                                    useSS = True
                                    break
                            # normal structure
                            if not useSS:
                                _packages.append(pName)
                            # delete since no needs anymore
                            del specialPackages[pName]
                            break
        # check glue packages
        for pName in gluePackages:
            if pName not in _packages:
                if os.path.isdir(_workArea+'/'+pName):
                    # check structured style
                    tmpDirList = os.listdir(_workArea+'/'+pName)
                    useSS = False
                    for tmpDir in tmpDirList:
                        if re.search('-\d+-\d+-\d+$',tmpDir) is not None:
                            fullPName = pName+'/'+tmpDir
                            if fullPName not in _packages:
                                _packages.append(fullPName)
                            useSS = True
                            break
                    # normal structure
                    if not useSS:
                        _packages.append(pName)
                else:
                    tmpLog.warning('glue package %s not found under %s' % (pName,_workArea))
        # return 
        return _packages


    # archive files
    def archiveFiles(_workArea,_packages,_archiveFullName):
        excludePattern = '.svn'
        for tmpPatt in excludeFile:
            # reverse regexp change
            tmpPatt = tmpPatt.replace('.*', '*')
            tmpPatt = tmpPatt.replace('\.', '.')
            excludePattern += " --exclude '%s'" % tmpPatt
        _curdir = os.getcwd()
        # change dir
        os.chdir(_workArea)
        for pack in _packages:
            # archive subdirs
            list = os.listdir(pack)
            for item in list:
                # ignore libraries
                if item.startswith('i686') or item.startswith('i386') or item.startswith('x86_64') \
                       or item=='pool' or item =='pool_plugins' or item == 'doc' or item == '.svn':
                    continue
                # check exclude files
                excludeFileFlag = False
                for tmpPatt in excludeFile:
                    if re.search(tmpPatt,'%s/%s' % (pack,item)) is not None:
                        excludeFileFlag = True
                        break
                if excludeFileFlag:
                    continue
                # run dir
                if item=='run':
                    files = []
                    getFileList('%s/%s/run' % (_workArea,pack),files,False)
                    # not resolve symlink (appending instead of replacing for backward compatibility)
                    tmpFiles = []
                    getFileList('%s/%s/run' % (_workArea,pack),tmpFiles,False,False)
                    for tmpFile in tmpFiles:
                        if tmpFile not in files:
                            files.append(tmpFile)
                    for iFile in files:
                        # converted to real path
                        file = os.path.realpath(iFile)
                        # archive .py/.dat/.C files only
                        if matchExtFile(file):
                            # remove special characters                    
                            sString=re.sub('[\+]','.',os.path.realpath(_workArea))
                            relPath = re.sub('^%s/' % sString, '', file)
                            # if replace is failed or the file is symlink, try non-converted path names
                            if relPath.startswith('/') or os.path.islink(iFile):
                                sString=re.sub('[\+]','.',workArea)
                                relPath = re.sub(sString+'/','',iFile)
                            if os.path.islink(iFile):
                                cmd = "tar -rh '%s' -f '%s' --exclude '%s'" % (relPath,_archiveFullName,excludePattern)
                                out = commands_get_output(cmd)
                            else:
                                cmd = "tar rf '%s' '%s' --exclude '%s'" % (_archiveFullName,relPath,excludePattern)
                                out = commands_get_output(cmd)
                            if verbose:
                                print(relPath)
                                if out != '':    
                                    print(out)
                    continue
                # else
                if dereferenceSymLinks:
                    cmd = "tar rfh '%s' '%s/%s' --exclude '%s'" % (_archiveFullName,pack,item,excludePattern)
                else:
                    cmd = "tar rf '%s' '%s/%s' --exclude '%s'" % (_archiveFullName,pack,item,excludePattern)
                out = commands_get_output(cmd)
                if verbose:
                    print("%s/%s" % (pack,item))
                    if out != '':    
                        print(out)
        # back to previous dir
        os.chdir(_curdir)

    #####################################################################
    # execute

    # get packages in private area 
    packages = getPackages(workArea,gluePackages)
    # check TestRelease since it doesn't create any links in InstallArea
    if os.path.exists('%s/TestRelease' % workArea):
        # the TestRelease could be created by hand
        packages.append('TestRelease')

    if verbose:
        tmpLog.debug("== private packages ==")
        for pack in packages:
            print(pack)
        tmpLog.debug("== private files ==")

    # create archive
    if archiveName == '':
        archiveName = 'sources.%s.tar' % MiscUtils.wrappedUuidGen()
    archiveFullName = "%s/%s" % (tmpDir,archiveName)
    # archive private area
    archiveFiles(workArea,packages,archiveFullName)
    # archive current (run) dir
    files = []
    os.chdir(workArea)
    getFileList('%s/%s' % (workArea,runDir),files,False,False)
    for file in files:
        # remove special characters                    
        sString=re.sub('[\+]','.',os.path.realpath(workArea))
        relPath = re.sub(sString+'/','',os.path.realpath(file))
        # if replace is failed or the file is symlink, try non-converted path names
        if relPath.startswith('/') or os.path.islink(file):
            sString=re.sub('[\+]','.',workArea)
            relPath = re.sub(sString+'/','',file)
        # archive .py/.dat/.C/.xml files only
        if not matchExtFile(relPath):
            continue
        # ignore InstallArea
        if relPath.startswith('InstallArea'):
            continue
        # check special files
        spBaseName = relPath
        if re.search('/',spBaseName) is not None:
            spBaseName = spBaseName.split('/')[-1]
        if spBaseName in specialFilesForAthena:
            warStr  = '%s in the current dir is sent to remote WNs, which might cause a database problem. ' % spBaseName
            warStr += 'If this is intentional please ignore this WARNING'
            tmpLog.warning(warStr)
        # check if already archived
        alreadyFlag = False
        for pack in packages:
            if relPath.startswith(pack):
                alreadyFlag = True
                break
        # archive
        if not alreadyFlag:
            if os.path.islink(file):
                out = commands_get_output("tar -rh '%s' -f '%s'" % (relPath, archiveFullName))
            else:
                out = commands_get_output("tar rf '%s' '%s'" % (archiveFullName, relPath))
            if verbose:
                print(relPath)
                if out != '':    
                    print(out)
    # back to current dir            
    os.chdir(currentDir)
    # return
    return archiveName,archiveFullName
    

# archive jobO files
def archiveJobOFiles(workArea,runDir,currentDir,tmpDir,verbose,archiveName=''):
    # archive jobO files    
    tmpLog = PLogger.getPandaLogger()
    tmpLog.info('archiving jobOs and modules')

    # get real jobOs
    def getJobOs(dir,files):
        list = os.listdir(dir)
        for item in list:
            if item in ['_CPack_Packages']:
                continue
            fullName=dir+'/'+item
            if os.path.isdir(fullName):
                # skip symlinks in include since they cause full scan on releases
                if os.path.islink(fullName) and re.search('InstallArea/include$',dir) is not None:
                    continue
                # dir
                getJobOs(fullName,files)
            else:
                # python and other extFiles
                if matchExtFile(fullName):
                    files.append(fullName)

    # get jobOs
    files = []
    os.chdir(workArea)
    getJobOs('%s' % workArea,files)
    # create archive
    if archiveName == '':
        archiveName = 'jobO.%s.tar' % MiscUtils.wrappedUuidGen()
    archiveFullName = "%s/%s" % (tmpDir,archiveName)
    # archive
    if verbose:
        tmpLog.debug("== py files ==")
    for file in files:
        # remove special characters                    
        sString=re.sub('[\+]','.',workArea)
        relPath = re.sub(sString+'/','',file)
        # append
        out = commands_get_output("tar -rh '%s' -f '%s'" % (relPath,archiveFullName))
        if verbose:
            print(relPath)
            if out != '':    
                print(out)
    # return
    return archiveName,archiveFullName


# archive InstallArea
def archiveInstallArea(workArea,groupArea,archiveName,archiveFullName,
                       tmpDir,nobuild,verbose):
    # archive jobO files    
    tmpLog = PLogger.getPandaLogger()
    tmpLog.info('archiving InstallArea')

    # get file list
    def getFiles(dir,files,ignoreLib,ignoreSymLink):
        if verbose:
            tmpLog.debug("  getFiles(%s)" % dir)
        try:    
            list = os.listdir(dir)
        except Exception:
            return
        for item in list:
            if ignoreLib and (item.startswith('i686') or item.startswith('i386') or
                              item.startswith('x86_64')):
                continue
            fullName=dir+'/'+item
            if os.path.isdir(fullName):
                # ignore symlinked dir just under InstallArea/include
                if ignoreSymLink and os.path.islink(fullName) and re.search('InstallArea/include$',dir) is not None:
                    continue
                # dir
                getFiles(fullName,files,False,ignoreSymLink)
            else:
                files.append(fullName)

    # get cmt files
    def getCMTFiles(dir,files):
        list = os.listdir(dir)
        for item in list:
            fullName=dir+'/'+item
            if os.path.isdir(fullName):
                # dir
                getCMTFiles(fullName,files)
            else:
                if re.search('cmt/requirements$',fullName) is not None:
                    files.append(fullName)

    # get files
    areaList = []
    # workArea must be first
    areaList.append(workArea)
    if groupArea != '':
        areaList.append(groupArea)
    # groupArea archive    
    groupFileName = re.sub('^sources','groupArea',archiveName)
    groupFullName = "%s/%s" % (tmpDir,groupFileName)
    allFiles = []    
    for areaName in areaList:    
        # archive
        if verbose:
            tmpLog.debug("== InstallArea under %s ==" % areaName)
        files = []
        cmtFiles = []
        os.chdir(areaName)
        if areaName==workArea:
            if not nobuild:
                # ignore i686 and include for workArea
                getFiles('InstallArea',files,True,True)
            else:
                # ignore include for workArea
                getFiles('InstallArea',files,False,True)
        else:
            # groupArea
            if not os.path.exists('InstallArea'):
                if verbose:
                    print("  Doesn't exist. Skip")
                continue
            getFiles('InstallArea',files,False,False)
            # cmt/requirements is needed for non-release packages
            for itemDir in os.listdir(areaName):
                if itemDir != 'InstallArea' and os.path.isdir(itemDir) and \
                       (not os.path.islink(itemDir)):
                    getCMTFiles(itemDir,cmtFiles)
        # remove special characters                    
        sString=re.sub('[\+]','.',os.path.realpath(areaName))
        # archive files if they are under the area
        for file in files+cmtFiles:
            relPath = re.sub(sString+'/','',os.path.realpath(file))
            # check exclude files
            excludeFileFlag = False
            for tmpPatt in excludeFile:
                if re.search(tmpPatt,relPath) is not None:
                    excludeFileFlag = True
                    break
            if excludeFileFlag:
                continue
            if not relPath.startswith('/'):
                # use files in private InstallArea instead of group InstallArea
                if file not in allFiles:
                    # append
                    if file in files:
                        out = commands_get_output("tar -rh '%s' -f '%s'" % (file,archiveFullName))
                    else:
                        # requirements files
                        out = commands_get_output("tar -rh '%s' -f '%s'" % (file,groupFullName))
                    allFiles.append(file)
                    if verbose:
                        print(file)
                        if out != '':    
                            print(out)
    # append groupArea to sources
    if groupArea != '' and (not nobuild):
        os.chdir(tmpDir)
        if os.path.exists(groupFileName):
            out = commands_get_output("tar -rh '%s' -f '%s'" % (groupFileName,archiveFullName))
            if out != '':    
                print(out)
            commands_get_output('rm -rf %s' % groupFullName)



# archive with cpack
def archiveWithCpack(withSource,tmpDir,verbose):
    tmpLog = PLogger.getPandaLogger()
    # define archive name
    if withSource:
        tmpLog.info('archiving source files')
        archiveName = 'sources.%s' % MiscUtils.wrappedUuidGen()
    else:
        tmpLog.info('archiving jobOs and modules')
        archiveName = 'jobO.%s' % MiscUtils.wrappedUuidGen()
    archiveFullName = "%s/%s" % (tmpDir,archiveName)
    # extract build dir
    buildDir = os.environ['CMAKE_PREFIX_PATH']
    buildDir = os.path.dirname(buildDir.split(':')[0])
    _curdir = os.getcwd()
    os.chdir(buildDir)
    tmpLog.info('The build directory is {0}'.format(buildDir))
    comStr = 'cpack -B {0} -D CPACK_PACKAGE_FILE_NAME={1} -G TGZ '.format(tmpDir,archiveName)
    comStr += '-D CPACK_PACKAGE_NAME="" -D CPACK_PACKAGE_VERSION="" -D CPACK_PACKAGE_VERSION_MAJOR="" '
    comStr += '-D CPACK_PACKAGE_VERSION_MINOR="" -D CPACK_PACKAGE_VERSION_PATCH="" '
    comStr += '-D CPACK_INSTALL_CMAKE_PROJECTS="" -D CPACK_PACKAGE_DESCRIPTION="" '
    if verbose:
        tmpLog.debug(comStr)
        comStr += '-V '
    tmpStat, tmpOut = commands_get_status_output_with_env(comStr)
    if tmpStat != 0:
        tmpLog.error('cpack failed')
        sys.exit(EC_Config)
    if verbose:
        print(tmpOut)
    archiveName += '.tar'
    archiveFullName += '.tar'
    os.chdir(tmpDir)
    comStr = 'tar xfz {0}.gz; tar cf {0} usr > /dev/null 2>&1; rm -rf usr _CPack_Packages {0}.gz'.format(archiveName)
    commands_get_output(comStr)
    os.chdir(_curdir)
    return archiveName,archiveFullName


# convert runConfig to outMap
def convertConfToOutput(runConfig,extOutFile,original_outDS,destination='',spaceToken='',descriptionInLFN='',
                        allowNoOutput=None):
    outMap = {}
    paramList = []
    # add IROOT
    if 'IROOT' not in outMap:
        outMap['IROOT'] = []
    # remove /
    outDSwoSlash = re.sub('/$','',original_outDS)
    outDsNameBase = outDSwoSlash
    tmpMatch = re.search('^([^\.]+)\.([^\.]+)\.',original_outDS)
    if tmpMatch is not None and original_outDS.endswith('/'):
        outDSwoSlash = '%s.%s' % (tmpMatch.group(1),tmpMatch.group(2))
        if descriptionInLFN != '':
            outDSwoSlash += descriptionInLFN
        outDSwoSlash += '.$JEDITASKID'
    # start conversion
    if runConfig.output.outNtuple:
        for sName in runConfig.output.outNtuple:
            lfn  = '%s.%s._${SN/P}.root' % (outDSwoSlash,sName)
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'ntuple' not in outMap:
                outMap['ntuple'] = []
            outMap['ntuple'].append((sName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outHist:
        lfn  = '%s.hist._${SN/P}.root' % outDSwoSlash
        tmpSuffix = '_HIST'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['hist'] = lfn
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outRDO:
        lfn  = '%s.RDO._${SN/P}.pool.root' % outDSwoSlash
        tmpSuffix = '_RDO'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['IROOT'].append((runConfig.output.outRDO,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)        
    if runConfig.output.outESD:
        lfn  = '%s.ESD._${SN/P}.pool.root' % outDSwoSlash
        tmpSuffix = '_ESD'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['IROOT'].append((runConfig.output.outESD,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outAOD:
        lfn  = '%s.AOD._${SN/P}.pool.root' % outDSwoSlash
        tmpSuffix = '_AOD'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['IROOT'].append((runConfig.output.outAOD,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outTAG:
        lfn  = '%s.TAG._${SN/P}.coll.root' % outDSwoSlash
        tmpSuffix = '_TAG'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['TAG'] = lfn
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outAANT:
        sNameList = []
        fsNameMap = {}
        for aName,sName,fName in runConfig.output.outAANT:
            # use first sName when multiple streams write to the same file
            realStreamName = sName
            if fName in fsNameMap:
                sName = fsNameMap[fName]
            else:
                fsNameMap[fName] = sName
            lfn  = '%s.%s._${SN/P}.root' % (outDSwoSlash,sName)       
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            if sName not in sNameList:
                sNameList.append(sName)
            if 'AANT' not in outMap:
                outMap['AANT'] = []
            outMap['AANT'].append((aName,realStreamName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outTHIST:
        for sName in runConfig.output.outTHIST:
            lfn  = '%s.%s._${SN/P}.root' % (outDSwoSlash,sName)
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'THIST' not in outMap:
                outMap['THIST'] = []
            outMap['THIST'].append((sName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outIROOT:
        for sIndex,sName in enumerate(runConfig.output.outIROOT):
            lfn  = '%s.iROOT%s._${SN/P}.%s' % (outDSwoSlash,sIndex,sName)
            tmpSuffix = '_iROOT%s' % sIndex
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'IROOT' not in outMap:
                outMap['IROOT'] = []
            outMap['IROOT'].append((sName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if extOutFile:
        for sIndex,sName in enumerate(extOutFile):
            # change * to X and add .tgz
            origSName = sName
            if sName.find('*') != -1:
                sName = sName.replace('*','XYZ')
                sName = '%s.tgz' % sName
            tmpExtStreamName = getExtendedExtStreamName(sIndex,sName,False)
            lfn  = '%s.%s._${SN/P}.%s' % (outDSwoSlash,tmpExtStreamName,sName)
            tmpSuffix = '_%s' % tmpExtStreamName
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'IROOT' not in outMap:
                outMap['IROOT'] = []
            outMap['IROOT'].append((origSName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outTAGX:
        for sName,oName in runConfig.output.outTAGX:
            lfn  = '%s.%s._${SN/P}.%s' % (outDSwoSlash,sName,oName)
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'IROOT' not in outMap:
                outMap['IROOT'] = []
            outMap['IROOT'].append((oName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outStream1:
        lfn  = '%s.Stream1._${SN/P}.pool.root' % outDSwoSlash
        tmpSuffix = '_Stream1'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['IROOT'].append((runConfig.output.outStream1,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)                
    if runConfig.output.outStream2:
        lfn  = '%s.Stream2._${SN/P}.pool.root' % outDSwoSlash
        tmpSuffix = '_Stream2'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['IROOT'].append((runConfig.output.outStream2,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outBS:
        lfn  = '%s.BS._${SN/P}.data' % outDSwoSlash
        tmpSuffix = '_BS'
        dataset = outDsNameBase + tmpSuffix + '/'
        outMap['BS'] = lfn
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outSelBS:
        lfn  = '%s.%s._${SN/P}.data' % (outDSwoSlash,runConfig.output.outSelBS)
        tmpSuffix = '_SelBS'
        dataset = outDsNameBase + tmpSuffix + '/'
        if 'IROOT' not in outMap:
            outMap['IROOT'] = []
        outMap['IROOT'].append(('%s.*.data' % runConfig.output.outSelBS,lfn))
        paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outStreamG:
        for sName,sOrigFileName in runConfig.output.outStreamG:
            lfn  = '%s.%s._${SN/P}.pool.root' % (outDSwoSlash,sName)
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            outMap['IROOT'].append((sOrigFileName,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)                
    if runConfig.output.outMeta:
        iMeta = 0
        for sName,sAsso in runConfig.output.outMeta:
            foundLFN = ''
            if sAsso == 'None':
                # non-associated metadata
                lfn  = '%s.META%s._${SN/P}.root' % (outDSwoSlash,iMeta)
                tmpSuffix = '_META%s' % iMeta
                dataset = outDsNameBase + tmpSuffix + '/'
                paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
                iMeta += 1
                foundLFN = lfn
            elif sAsso in outMap:
                # Stream1,2
                foundLFN = outMap[sAsso]
            elif sAsso in ['StreamESD','StreamAOD']:
                # ESD,AOD
                stKey = re.sub('^Stream','',sAsso)
                if stKey in outMap:
                    foundLFN = outMap[stKey]
                else:
                    # check StreamG when ESD/AOD are not defined as algorithms
                    if 'StreamG' in outMap:
                        for tmpStName,tmpLFN in outMap['StreamG']:
                            if tmpStName == sAsso:
                                foundLFN = tmpLFN
            elif sAsso == 'StreamRDO' and 'StreamRDO' in outMap:
                # RDO
                stKey = re.sub('^Stream','',sAsso)
                if stKey in outMap:
                    foundLFN = outMap[stKey]
            else:
                # general stream
                if 'StreamG' in outMap:
                    for tmpStName,tmpLFN in outMap['StreamG']:
                        if tmpStName == sAsso:
                            foundLFN = tmpLFN
            if foundLFN != '':
                if 'Meta' not in outMap:
                    outMap['Meta'] = []
                outMap['Meta'].append((sName,foundLFN))
    if runConfig.output.outMS:
        for sName,sAsso in runConfig.output.outMS:
            lfn  = '%s.%s._${SN/P}.pool.root' % (outDSwoSlash,sName)
            tmpSuffix = '_%s' % sName
            dataset = outDsNameBase + tmpSuffix + '/'
            if 'IROOT' not in outMap:
                outMap['IROOT'] = []
            outMap['IROOT'].append((sAsso,lfn))
            paramList += MiscUtils.makeJediJobParam(lfn,dataset,'output',hidden=True,allowNoOutput=allowNoOutput)
    if runConfig.output.outUserData:
        for sAsso in runConfig.output.outUserData:
            # look for associated LFN
            foundLFN = ''            
            if sAsso in outMap:
                # Stream1,2
                foundLFN = outMap[sAsso]
            elif sAsso in ['StreamRDO','StreamESD','StreamAOD']:
                # RDO,ESD,AOD
                stKey = re.sub('^Stream','',sAsso)
                if stKey in outMap:
                    foundLFN = outMap[stKey]
            else:
                # general stream
                if 'StreamG' in outMap:
                    for tmpStName,tmpLFN in outMap['StreamG']:
                        if tmpStName == sAsso:
                            foundLFN = tmpLFN
            if foundLFN != '':
                if 'UserData' not in outMap:
                    outMap['UserData'] = []
                outMap['UserData'].append(foundLFN)
    # remove IROOT if unnecessary
    if 'IROOT' in outMap and outMap['IROOT'] == []:
        del outMap['IROOT'] 
    # set destination
    if destination != '':
        for tmpParam in paramList:
            tmpParam['destination'] = destination
    # set token
    if spaceToken != '':
        for tmpParam in paramList:
            tmpParam['token'] = spaceToken
    # return
    return outMap,paramList


# get CMTCONFIG + IMG
def getCmtConfigImg(athenaVer=None, cacheVer=None, nightVer=None, cmtConfig=None, verbose=False, architecture=None):
    # get CMTCONFIG
    retVal = ''
    if architecture and '@' in architecture:
        retVal = architecture.split('@')[0]
    if not retVal:
        retVal = getCmtConfig(athenaVer, cacheVer, nightVer, cmtConfig, verbose)
    # get base platform + HW specs
    if architecture and '@' in architecture:
        # architecture with base platform
        spec_str = '@' + architecture.split('@')[-1]
    elif 'ALRB_USER_PLATFORM' in os.environ:
        # base platform from ALRB
        spec_str = '@' + os.environ['ALRB_USER_PLATFORM']
        # HW specs
        if architecture:
            spec_str += architecture
    else:
        # architecture w/o base platform or even empty architecture
        spec_str = architecture
    # append base platform + HW specs if any
    if spec_str:
        if retVal is None:
            retVal = ''
        retVal = retVal + spec_str
    return retVal


# get CMTCONFIG
def getCmtConfig(athenaVer=None,cacheVer=None,nightVer=None,cmtConfig=None,verbose=False):
    # use user-specified cmtconfig
    if cmtConfig:
        return cmtConfig
    # local settting
    if 'CMTCONFIG' in os.environ:
        return os.environ['CMTCONFIG']
    # undefined in Athena
    if athenaVer or cacheVer:
        # get logger
        tmpLog = PLogger.getPandaLogger()
        tmpLog.error('environment variable CMTCONFIG is undefined. Please set --cmtConfig')
        sys.exit(EC_Config)
    return None
    

# check CMTCONFIG
def checkCmtConfig(localCmtConfig,userCmtConfig,noBuild):
    # didn't specify CMTCONFIG
    if userCmtConfig in ['',None]:
        return True
    # CVMFS version format
    if re.search('-gcc\d+\.\d+$',userCmtConfig) is not None:
        return True
    # get logger
    tmpLog = PLogger.getPandaLogger()
    # CMTCONFIG is undefined locally
    if localCmtConfig in ['',None]:
        return True
    # user-specified CMTCONFIG is inconsitent with local CMTCONFIG
    if userCmtConfig != localCmtConfig and noBuild:
        errStr  = 'You cannot use --noBuild when --cmtConfig=%s is inconsistent with local CMTCONFIG=%s ' % (userCmtConfig,localCmtConfig)
        errStr += 'since you need re-compile source files on remote worker-node. Please remove --noBuild'
        tmpLog.error(errStr)
        return False
    # return OK
    return True


# use AMI for AutoConf
def getJobOtoUseAmiForAutoConf(inDS,tmpDir):
    # no input
    if inDS == '':
        return ''
    # use first one
    amiURL = 'ami://%s' % inDS.split(',')[0]
    # remove /
    if amiURL.endswith('/'):
        amiURL = amiURL[:-1]
    inputFiles = [amiURL]
    # create jobO fragment
    optFileName = tmpDir + '/' + MiscUtils.wrappedUuidGen() + '.py'
    oFile = open(optFileName,'w')
    oFile.write("""
try:
    import AthenaCommon.AthenaCommonFlags
    
    def _dummyFilesInput(*argv):
        return %s

    AthenaCommon.AthenaCommonFlags.FilesInput.__call__ = _dummyFilesInput
except Exception:
    pass

try:
    import AthenaCommon.AthenaCommonFlags
    
    def _dummyGet_Value(*argv):
        return %s

    for tmpAttr in dir (AthenaCommon.AthenaCommonFlags):
        import re
        if re.search('^(Pool|BS).*Input$',tmpAttr) is not None:
            try:
                getattr(AthenaCommon.AthenaCommonFlags,tmpAttr).get_Value = _dummyGet_Value
            except Exception:
                pass
except Exception:
    pass
""" % (inputFiles,inputFiles))
    oFile.close()
    # reutrn file name
    return optFileName


# use CMake
def useCMake():
    return 'CMAKE_PREFIX_PATH' in os.environ


# parse athenaTag
def parse_athena_tag(athena_tag, verbose, tmp_log):
    athenaVer = ''
    cacheVer  = ''
    nightVer  = ''
    # loop over all tags
    items = athena_tag.split(',')
    usingNightlies = False
    for item in items:
        # releases
        match = re.search(r'^(\d+\.\d+\.\d+)', item)
        if match:
            athenaVer = 'Atlas-%s' % match.group(1)
            # cache
            cmatch = re.search(r'^(\d+\.\d+\.\d+\.\d+\.*\d*)$',item)
            if cmatch is not None:
                cacheVer += '_%s' % cmatch.group(1)
            else:
                cacheVer += '_%s' % match.group(1)
            continue
        # nightlies
        match = re.search(r'^(\d+\.\d+\.X|\d+\.X\.\d+)$',item)
        if match:
            athenaVer = 'Atlas-%s' % match.group(1)
            continue
        # master or XX.YY
        match = re.search(r'^\d+\.\d+$', item)
        if item.startswith('master') or match:
            athenaVer = 'Atlas-%s' % item
            continue
        # old nightlies
        if item.startswith('rel_'):
            usingNightlies = True
            if 'dev' in items:
                athenaVer = 'Atlas-dev'
            elif 'devval' in items:
                athenaVer = 'Atlas-devval'
            cacheVer  = '-AtlasOffline_%s' % item
            continue
        # nightlies
        if item in ['latest', 'r27'] or re.search(r'^\d{4}-\d{2}-\d{2}T\d{4}$', item):
            cacheVer += '_%s' % item
            continue
        # CMTCONFIG
        if item in ['32', '64']:
            tmp_log.error("%s in --athenaTag is unsupported. Please use --cmtConfig instead" % item)
            sys.exit(EC_Config)
        # ignoring AtlasOffline
        if item in ['AtlasOffline']:
            continue
        # regarded as project
        cacheVer = '-'+item+cacheVer
    # check cache
    if re.search(r'^-.+_.+$', cacheVer) is None:
        if re.search(r'^_\d+\.\d+\.\d+\.\d+$', cacheVer) is not None:
            # use AtlasProduction
            cacheVer = '-AtlasProduction'+cacheVer
        elif 'AthAnalysisBase' in cacheVer or 'AthAnalysis' in cacheVer:
            # AthAnalysis
            cacheVer  = cacheVer + '_%s' % athenaVer
            athenaVer = ''
        else:
            # unknown
            cacheVer = ''
    # use dev nightlies
    if usingNightlies and athenaVer == '':
        athenaVer = 'Atlas-dev'
    return athenaVer, cacheVer, nightVer



import re
import os
import gtk
import pango
import gobject
import gtk.glade
import threading

try:
    import Queue
except ImportError:
    import queue as Queue
from . import PLogger
from . import LocalJobSpec
from .MiscUtils import commands_get_output, commands_get_status_output
try:
    long()
except Exception:
    long = int

gobject.threads_init()
gtk.gdk.threads_init()


  
# thread to synchronize database in background
class Synchronizer(threading.Thread):
    
    # constructor
    def __init__(self,syncQueue,pEmitter):
        # init thread
        threading.Thread.__init__(self)
        # queue
        self.syncQueue = syncQueue
        # try to get core
        try:
            self.pbookCore = self.syncQueue.get_nowait()
        except Exception:
            self.pbookCore = None
        # emitter
        self.pEmitter = pEmitter

    # run
    def run(self):
        if self.pbookCore is None:
            return
        # synchronize database
        self.pbookCore.sync()
        # put back queue
        self.syncQueue.put(self.pbookCore)
        # emit signal
        gobject.idle_add(self.pEmitter.emit,"on_syncEnd")



# thread to retry in background
class RetryWorker(threading.Thread):
    
    # constructor
    def __init__(self,jobID,guiGlobal,retryQueue,pEmitter):
        # init thread
        threading.Thread.__init__(self)
        # JobID
        self.jobID = jobID
        # global data
        self.guiGlobal = guiGlobal
        # queue
        self.retryQueue = retryQueue
        # try to get core
        try:
            self.pbookCore = self.retryQueue.get_nowait()
        except Exception:
            self.pbookCore = None
        # emitter
        self.pEmitter = pEmitter

    # run
    def run(self):
        if self.pbookCore is None:
            return
        # retry
        self.pbookCore.retry(long(self.jobID))
        # put back queue
        self.retryQueue.put(self.pbookCore)
        # reset offset
        self.guiGlobal.resetJobOffset()
        # emit signal
        gobject.idle_add(self.pEmitter.emit,"on_syncEnd")




# thread to open url
class UrlOpener(threading.Thread):
    
    # constructor
    def __init__(self,url,queue):
        # init thread
        threading.Thread.__init__(self)
        # url
        self.url = url
        # queue
        self.queue = queue
        # browser type
        self.browser = 'firefox'
        # logger
        self.tmpLog = PLogger.getPandaLogger()


    # run
    def run(self):
        if self.browser == 'firefox':
            # check application
            status,output = commands_get_status_output('which %s' % self.browser)
            if status != 0:
                self.tmpLog.error('%s is unavailable' % self.browser)
                return
            # check version
            output = commands_get_output('%s -v' % self.browser)
            version = output.split()[2]
            version = version[:-1]
            if version < '2.0':
                self.tmpLog.warning("too old %s : version %s It wouldn't work properly" % (self.browser,version))
            # open url
            com = '%s %s' % (self.browser,self.url)
            commands_get_output(com)
        # release queue
        self.queue.put(True)
        return

        

# global data
class PBookGuiGlobal:

    # constructor
    def __init__(self):
        # list of jobs in local DB
        self.jobMap = {}
        # current JobID
        self.currentJob = None
        # current job offset
        self.jobOffset = 0
        # lock
        self.lock = Queue.Queue(1)
        self.lock.put(True)
        

    # set job map
    def setJobMap(self,jobMap):
        lock = self.lock.get()
        self.jobMap = jobMap
        self.lock.put(lock)


    # get job map
    def getJobMap(self):
        return self.jobMap


    # set current job
    def setCurrentJob(self,jobID):
        lock = self.lock.get()        
        if jobID in self.jobMap:
            self.currentJob = jobID
        self.lock.put(lock)
        

    # update job
    def updateJob(self,job):
        lock = self.lock.get()
        if isinstance(job,LocalJobSpec.LocalJobSpec):
            self.jobMap[job.JobID] = job
        else:
            self.jobMap[job.JobsetID] = job            
        self.lock.put(lock)
        

    # get current job
    def getCurrentJob(self):
        return self.currentJob


    # get job
    def getJob(self,jobID):
        return self.jobMap[jobID]


    # reset offset of jobs
    def resetJobOffset(self):
        lock = self.lock.get()
        # reset to 0
        self.jobOffset = 0
        self.lock.put(lock)
        
        
    # set offset of jobs
    def setJobOffset(self,change):
        lock = self.lock.get()        
        # try
        tmpOffset = self.jobOffset + change
        # reset if out of range
        if tmpOffset >= len(self.jobMap):
            tmpOffset = self.jobOffset
        elif tmpOffset < 0:
            tmpOffset = 0
        # set
        self.jobOffset = tmpOffset
        self.lock.put(lock)
        

    # get offset
    def getJobOffset(self):
        return self.jobOffset



# jump to JobID
class PJumper:

    # constructor
    def __init__(self,guiGlobal,pEmitter):
        # jobID to jump to
        self.jobID = None
        # global data
        self.guiGlobal = guiGlobal
        # emitter
        self.pEmitter = pEmitter


    # set jobID
    def setJobID(self,jobID):
        self.jobID = jobID

        
    # action
    def on_clicked(self,tag,textview,event,iter):
        # mouse clicked
        if event.type == gtk.gdk.BUTTON_PRESS:
            if self.jobID is not None:
                # set jobID to global
                self.guiGlobal.setCurrentJob(self.jobID)
                # emit
                self.pEmitter.emit("on_setNewJob")



# text view for summary
class PSumView:
    
    # constructor
    def __init__(self,sumView,guiGlobal,pEmitter):
        # widget
        self.sumView = sumView
        # global data
        self.guiGlobal = guiGlobal
        # emitter
        self.pEmitter = pEmitter
        # jumper
        self.jumper = {'retryID'      : PJumper(self.guiGlobal,self.pEmitter),
                       'provenanceID' : PJumper(self.guiGlobal,self.pEmitter),
                       'retrySetID'   : PJumper(self.guiGlobal,self.pEmitter),
                       'parentSetID'  : PJumper(self.guiGlobal,self.pEmitter),                       
                       }
        self.firstJump = {}
        for tmpJumpName in self.jumper.keys():
            self.firstJump[tmpJumpName] = True
        # sizes
        self.nLines = 20+1
        self.nColumns = 4
        # resize
        self.sumView.resize(self.nLines,self.nColumns)
        # create TextViews
        self.allBufList  = []
        self.textViewMap = {}
        for iLine in range(self.nLines):
            bufList = []
            for item in ('label','value'):
                # text buffer
                textBuf = gtk.TextBuffer()
                # set tag
                tag = textBuf.create_tag('default')
                tag.set_property("font", "monospace")
                tag = textBuf.create_tag('red')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','red2')
                tag = textBuf.create_tag('pink')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','deeppink')
                tag = textBuf.create_tag('green')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','green4')
                tag = textBuf.create_tag('yellow')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','darkgoldenrod')
                tag = textBuf.create_tag('navy')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','navy')
                tag = textBuf.create_tag('skyblue')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','darkturquoise')
                tag = textBuf.create_tag('purple')
                tag.set_property("font", "monospace")
                tag.set_property('foreground','blueviolet')
                # create textview
                textView = gtk.TextView(textBuf)
                self.textViewMap[textBuf] = textView
                # properties
                textView.set_editable(False)
                textView.set_cursor_visible(False)
                # set size and justification
                if item == 'label':
                    textView.set_size_request(120,-1)
                    textView.set_justification(gtk.JUSTIFY_RIGHT)
                else:
                    textView.set_size_request(460,-1)
                    textView.set_justification(gtk.JUSTIFY_LEFT)
                    textView.set_right_margin(20)
                # color
                textView.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("ghostwhite"))                
                # wrap mode
                textView.set_wrap_mode(gtk.WRAP_CHAR)
                # append
                if item == 'label':
                    self.sumView.attach(textView,0,1,iLine,iLine+1)
                else:
                    self.sumView.attach(textView,1,self.nColumns-1,iLine,iLine+1)
                bufList.append(textBuf)
            # append
            self.allBufList.append(bufList)
        # show
        self.sumView.show_all()
        # cursors
        self.cursors = {'normal' : gtk.gdk.Cursor(gtk.gdk.XTERM),
                        'link'   : gtk.gdk.Cursor(gtk.gdk.HAND2)
                        }

        
    # show summary
    def showJobInfo(self,widget):
        # get job
        jobID = self.guiGlobal.getCurrentJob()
        job   = self.guiGlobal.getJob(jobID)
        # make string
        strJob  = "\n"
        strJob += str(job)
        strJob += "\n"       
        # split to lines
        lines = strJob.split('\n')
        # delte
        for tmpBufList in self.allBufList:
            for textbuf in tmpBufList:
                # delete
                textbuf.delete(textbuf.get_start_iter(),
                               textbuf.get_end_iter())
        # fill
        jobStatusRows  = False
        jobStatusLines = ''
        jobStatusIdx   = 0
        for iLine in range(len(lines)):
            # check limit
            if iLine+1 > self.nLines:
                continue
            # decompose line to label and value
            line = lines[iLine]
            match = re.search('^([^:]+:)(.*)',line)
            if match is None:
                items = [line,'']
            else:
                items = match.groups()
            # fill
            if not jobStatusRows:
                iItem = 0
                for strLabel in items:
                    # remove redundant white spaces
                    strLabel = strLabel.strip()
                    strLabel += ' '
                    # get textbuffer
                    textbuf = self.allBufList[iLine][iItem]
                    # delete
                    textbuf.delete(textbuf.get_start_iter(),
                                   textbuf.get_end_iter())
                    # set color
                    tagname = 'default'
                    if (line.strip().startswith('jobStatus') or line.strip().startswith('status')) and iItem != 0:
                        if strLabel.find('frozen') != -1:
                            tagname = 'navy'
                        elif strLabel.find('killing') != -1:
                            tagname = 'pink'                            
                        else:
                            tagname = 'skyblue'
                    # add jumper
                    match = re.search('\s*(\S+)\s*:',line)
                    if match is not None:
                        realLabel = match.group(1)
                        if realLabel in self.jumper and iItem != 0:
                            # set jobID
                            jumper = self.jumper[realLabel]
                            jobID = strLabel.strip()
                            if jobID in ['0','']:
                                jumper.setJobID(None)
                                strLabel = ''
                                # change mouse cursor
                                self.textViewMap[textbuf].get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(self.cursors['normal'])
                            else:
                                jumper.setJobID(jobID)
                                # set tagname
                                tagname = 'hyperlink'
                                # change mouse cursor
                                self.textViewMap[textbuf].get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(self.cursors['link'])
                            # setup textview
                            if self.firstJump[realLabel]:
                                # add connection
                                tag = textbuf.create_tag("hyperlink",foreground='blue',
                                                         underline=pango.UNDERLINE_SINGLE)
                                tag.connect('event',jumper.on_clicked)
                                # disable first flag
                                self.firstJump[realLabel] = False
                    # write
                    textbuf.insert_with_tags_by_name(textbuf.get_end_iter(),
                                                     strLabel,tagname)
                    # increment
                    iItem += 1
                if line.strip().startswith('jobStatus') or line.strip().startswith('status'):
                    jobStatusRows = True
                    jobStatusIdx = iLine +1
                    try:
                        # get textbuffer
                        textbuf = self.allBufList[iLine+1][1]
                        # delete
                        textbuf.delete(textbuf.get_start_iter(),
                                       textbuf.get_end_iter())
                    except Exception:
                        pass
            else:
                # get textbuffer
                textbuf = self.allBufList[jobStatusIdx][1]
                # change color
                tagname = 'default'                
                if line.find('finished') != -1:
                    tagname = 'green'
                elif line.find('failed') != -1:
                    tagname = 'red'
                elif line.find('cancelled') != -1:
                    tagname = 'purple'
                else:
                    tagname = 'yellow'
                # reformat
                items = line.strip().split()
                if len(items) > 1:
                    strLine = '%10s : %s\n' % (items[0],items[-1])
                else:
                    strLine = '\n'
                textbuf.insert_with_tags_by_name(textbuf.get_end_iter(),
                                                 strLine,tagname)
        

# text view for status
class PStatView:

    # constructor
    def __init__(self,statView):
        # text view
        self.statView = statView
        # text buffer
        self.buffer = self.statView.get_buffer()
        # tags
        self.tags = {}
        # info
        tag = self.buffer.create_tag('INFO')
        tag.set_property("font", "monospace")
        #tag.set_property("size-points", 10)
        tag.set_property('foreground','blue')
        self.tags['INFO'] = tag
        # debug
        tag = self.buffer.create_tag('DEBUG')
        tag.set_property("font", "monospace")
        #tag.set_property("size-points", 10)
        tag.set_property('foreground','black')
        self.tags['DEBUG'] = tag
        # warning
        tag = self.buffer.create_tag('WARNING')
        tag.set_property("font", "monospace")
        #tag.set_property("size-points", 10)
        tag.set_property('foreground','orange')
        self.tags['WARNING'] = tag
        # error
        tag = self.buffer.create_tag('ERROR')
        tag.set_property("font", "monospace")
        #tag.set_property("size-points", 10)
        tag.set_property('foreground','red')
        self.tags['ERROR'] = tag
        # format
        self.format = ' %7s : %s\n'
        # dummy handlers
        self.handlers = ['']
        # set color
        self.statView.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("ghostwhite"))        
        # queue for serialization
        self.queue = Queue.Queue(1)
        self.queue.put(True)
        # wrap mode
        self.statView.set_wrap_mode(gtk.WRAP_CHAR)
        

    # formatter
    def formatter(self,level,msg):
        # format
        return self.format % (level,msg)


    # write message with level
    def write(self,level,msg):
        # insert message
        message = self.formatter(level,msg)
        self.buffer.insert_with_tags_by_name(self.buffer.get_end_iter(),
                                             message,level)
        # scroll
        mark = self.buffer.get_mark("insert")
        self.statView.scroll_mark_onscreen(mark)
        gtk.threads_leave()

        
    # emulation for logging
    def info(self,msg,withLock=False):
        sem = self.queue.get()
        if withLock:
            self.write('INFO',msg)
        else:
            gobject.idle_add(self.write,'INFO',msg)            
        self.queue.put(sem)


    # emulation for logging
    def debug(self,msg):
        sem = self.queue.get()        
        gobject.idle_add(self.write,'DEBUG',msg)        
        self.queue.put(sem)
        

    # emulation for logging
    def warning(self,msg):
        sem = self.queue.get()        
        gobject.idle_add(self.write,'WARNING',msg)
        self.queue.put(sem)
        

    # emulation for logging
    def error(self,msg):
        sem = self.queue.get()        
        gobject.idle_add(self.write,'ERROR',msg)        
        self.queue.put(sem)


# tree view for job list
class PTreeView:
    
    # constructor
    def __init__(self,treeView,guiGlobal,pbookCore,pEmitter):
        # tree view
        self.treeView = treeView
        # global data
        self.guiGlobal = guiGlobal
        # core of pbook
        self.pbookCore = pbookCore
        # emitter
        self.pEmitter = pEmitter
        # column names
        self.columnNames =  ['JobID','creationTime']
        # max number of jobs in tree
        self.maxJobs = 20
        # load job list
        self.reloadJobList()
        # set columns
        self.setColumns()
        # connect signals
        self.treeView.get_selection().connect('changed',self.showJob)
        # pixel buffer
        self.pbMap = {}
        icon_map = {'finished' : 'green.png',
                                    'failed'   : 'red.png',
                                    'running'  : 'yellow.png',
                                    'cancelled': 'orange.png',
                                   }
        for tmpStatus in icon_map:
            tmpFname = icon_map[tmpStatus]
            pixbuf = gtk.gdk.pixbuf_new_from_file(os.environ['PANDA_SYS'] \
                                                  + "/etc/panda/icons/" + tmpFname)
            self.pbMap[tmpStatus] = pixbuf
        # set color
        self.treeView.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("ghostwhite"))         

        
    # reload job list
    def reloadJobList(self,slot_obj=None,refresh=True):
        # read Jobs from DB
        if refresh:
            tmpList = self.pbookCore.getLocalJobList()
            # convert to map
            jobList = {}
            for tmpJob in tmpList:
                if isinstance(tmpJob,LocalJobSpec.LocalJobSpec):
                    jobList[tmpJob.JobID] = tmpJob
                else:
                    jobList[tmpJob.JobsetID] = tmpJob                    
            # set global data
            self.guiGlobal.setJobMap(jobList)
        else:
            jobList = self.guiGlobal.getJobMap()
        # make list sore
        listModel = gtk.ListStore(object)
        # sort jobs
        tmpJobIntoStrMap = {}
        for tmpJobIDStr in jobList.keys():
            tmpJobIntoStrMap[long(tmpJobIDStr)] = tmpJobIDStr
        tmpJobIDs = list(tmpJobIntoStrMap)
        tmpJobIDs.sort()
        tmpJobIDs.reverse()
        # get offset
        offset = self.guiGlobal.getJobOffset()
        # truncate
        jobIDs = []
        for tmpJobID in tmpJobIDs[offset:offset+self.maxJobs]:
            jobIDs.append(tmpJobIntoStrMap[tmpJobID])
        # append
        for jobID in jobIDs:
            listModel.append([jobID])
        # set model
        self.treeView.set_model(listModel)


    # set columns
    def setColumns(self):
        # JobID has icon+text
        cellpb = gtk.CellRendererPixbuf()
        tvcolumn = gtk.TreeViewColumn("  %s" % self.columnNames[0], cellpb)
        tvcolumn.set_cell_data_func(cellpb, self.setIcon)
        cell = gtk.CellRendererText()
        tvcolumn.pack_start(cell, False)
        tvcolumn.set_cell_data_func(cell, self.setJobID)
        self.treeView.append_column(tvcolumn)
        # text only for other parameters
        for attr in self.columnNames[1:]:
            cell = gtk.CellRendererText()
            tvcolumn = gtk.TreeViewColumn("     %s" % attr, cell)
            tvcolumn.set_cell_data_func(cell,self.setValue,attr)
            self.treeView.append_column(tvcolumn)


    # set ID column
    def setJobID(self,column,cell,model,iter):
        jobID = model.get_value(iter, 0)
        cell.set_property('text', jobID)


    # set icon
    def setIcon(self,column,cell,model,iter):
        jobID = model.get_value(iter, 0)
        job = self.guiGlobal.getJob(jobID)
        if job.dbStatus != 'frozen':
            pb = self.pbMap['running']
        elif job.jobStatus.find('failed') != -1:
            pb = self.pbMap['failed']
        elif job.jobStatus.find('cancelled') != -1:
            pb = self.pbMap['cancelled']
        else:
            pb = self.pbMap['finished']
        cell.set_property('pixbuf', pb)

        
    # set columns
    def setValue(self,column,cell,model,iter,attr):
        jobID = model.get_value(iter, 0)
        job = self.guiGlobal.getJob(jobID)
        var = getattr(job,attr)
        cell.set_property('text', "   %s" % var)  

        
    # show job info
    def showJob(self,selection):
        # get job info
        model,selected = selection.get_selected_rows()
        if len(selected) == 0:
            return
        iter = model.get_iter(selected[0])
        jobID = model.get_value(iter,0)
        self.guiGlobal.setCurrentJob(jobID)
        # emit signal
        self.pEmitter.emit("on_setNewJob")


# sync button
class PSyncButton:

    # constructor
    def __init__(self,syncButton,pbookCore,pEmitter):
        # sync button
        self.syncButton = syncButton
        # core
        self.pbookCore = pbookCore
        # queue for synchronizer
        self.syncQueue = Queue.Queue(1)
        self.syncQueue.put(pbookCore)
        # emitter
        self.pEmitter = pEmitter


    # synchronze database
    def on_clicked(self,widget):
        synchronizer = Synchronizer(self.syncQueue,self.pEmitter)
        synchronizer.start()



# range button
class PRangeButton:

    # constructor
    def __init__(self,rangeButton,guiGlobal,change,pEmitter):
        # button
        self.rangeButton = rangeButton
        # global
        self.guiGlobal = guiGlobal
        # changed value
        self.change = change
        # emitter
        self.pEmitter = pEmitter


    # clicked action
    def on_clicked(self,widget):
        # get current offset
        oldOffset = self.guiGlobal.getJobOffset()
        # chage offset
        self.guiGlobal.setJobOffset(self.change)
        # get new offset
        newOffset = self.guiGlobal.getJobOffset()
        # emit signal
        if oldOffset != newOffset:
            self.pEmitter.emit("on_rangeChanged",False)



# wev button
class PWebButton:

    # constructor
    def __init__(self,url):
        # url
        self.url = url
        # queue to avoid duplication
        self.queue = Queue.Queue(1)
        self.queue.put(True)


    # clicked action
    def on_clicked(self,widget):
        try:
            ret = self.queue.get_nowait()
        except Exception:
            return
        # make opener
        thr = UrlOpener(self.url,self.queue)
        thr.start()



# config button
class PConfigButton:

    # constructor
    def __init__(self,gladefile):
        # XML name
        self.gladefile = gladefile

    # clicked action
    def on_clicked(self,widget):
        return


# update button
class PUpdateButton:

    # constructor
    def __init__(self,pbookCore,guiGlobal,pEmitter):
        # core
        self.pbookCore = pbookCore
        # global
        self.guiGlobal = guiGlobal
        # emitter
        self.pEmitter = pEmitter


    # clicked action
    def on_clicked(self,widget):
        # logger
        tmpLog = PLogger.getPandaLogger()
        # get jobID
        jobID = self.guiGlobal.getCurrentJob()
        if jobID is None:
            tmpLog.warning('No job is selected. Please click a job in the left list first')
            return
        # skip if frozen
        job = self.guiGlobal.getJob(jobID)
        if job.dbStatus == 'frozen':
            tmpLog.info('Update is not required for frozen jobs')
            return
        # get updated info
        updatedJob = self.pbookCore.statusJobJobset(long(jobID))
        # update global data
        self.guiGlobal.updateJob(updatedJob)
        # emit signal
        self.pEmitter.emit("on_setNewJob")
        # update list
        self.pEmitter.emit("on_rangeChanged",False)
        


# kill button
class PKillButton:

    # constructor
    def __init__(self,pbookCore,guiGlobal,pEmitter):
        # core
        self.pbookCore = pbookCore
        # global
        self.guiGlobal = guiGlobal
        # emitter
        self.pEmitter = pEmitter


    # clicked action
    def on_clicked(self,widget):
        # logger
        tmpLog = PLogger.getPandaLogger()
        # get jobID
        jobID = self.guiGlobal.getCurrentJob()
        if jobID is None:
            tmpLog.warning('No job is selected. Please click a job in the left list first')
            return
        # skip if frozen
        job = self.guiGlobal.getJob(jobID)
        if job.dbStatus == 'frozen':
            tmpLog.info('All subJobs already finished/failed')            
            return
        # kill
        self.pbookCore.kill(long(jobID))
        # get job info
        updatedJob = self.pbookCore.getJobJobsetInfo(jobID)
        # update global data
        self.guiGlobal.updateJob(updatedJob)
        # emit signal
        self.pEmitter.emit("on_setNewJob")



# retry button
class PRetryButton:

    # constructor
    def __init__(self,pbookCore,guiGlobal,pEmitter):
        # core
        self.pbookCore = pbookCore
        # global
        self.guiGlobal = guiGlobal
        # emitter
        self.pEmitter = pEmitter
        # queue for retry workers
        self.retryQueue = Queue.Queue(1)
        self.retryQueue.put(pbookCore)


    # clicked action
    def on_clicked(self,widget):
        # logger
        tmpLog = PLogger.getPandaLogger()
        # get jobID
        jobID = self.guiGlobal.getCurrentJob()
        if jobID is None:
            tmpLog.warning('No job is selected. Please click a job in the left list first')
            return
        # skip if frozen
        job = self.guiGlobal.getJob(jobID)
        if job.dbStatus != 'frozen':
            tmpLog.warning('Cannot retry running jobs')
            return
        # retry
        thr = RetryWorker(jobID,self.guiGlobal,self.retryQueue,self.pEmitter)
        thr.start()



# emitter
class PEmitter(gobject.GObject):
    __gsignals__ = {
        'on_syncEnd'      : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
        'on_setNewJob'    : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
        'on_rangeChanged' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (bool,)),
        'on_triggerSync'  : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),        
        }
    # constructor
    def __init__(self):
        gobject.GObject.__init__(self)

# register emitter
gobject.type_register(PEmitter)



# wrapper for gtk.main_quit
def overallQuit(*arg):
    gtk.main_quit()
    commands_get_output('kill -- -%s' % os.getpgrp())



# main
class PBookGuiMain:
    
    # constructor
    def __init__(self,pbookCore):

        # load glade file
        gladefile = os.environ['PANDA_SYS'] + "/etc/panda/pbook.glade"  
        mainWindow = gtk.glade.XML(gladefile,"mainWindow")
        # global data
        pbookGuiGlobal = PBookGuiGlobal()
        # emitter
        self.pEmitter = PEmitter()
        # instantiate components
        statView = PStatView(mainWindow.get_widget("statusTextView"))
        sumView  = PSumView(mainWindow.get_widget("summaryTable"),
                            pbookGuiGlobal,self.pEmitter)
        treeView = PTreeView(mainWindow.get_widget("mainTreeView"),
                             pbookGuiGlobal,pbookCore,self.pEmitter)
        syncButton     = PSyncButton(mainWindow.get_widget("syncButton"),
                                     pbookCore,self.pEmitter)
        backButton     = PRangeButton(mainWindow.get_widget("backButton"),
                                      pbookGuiGlobal,-treeView.maxJobs,self.pEmitter)
        forwardButton  = PRangeButton(mainWindow.get_widget("forwardButton"),
                                      pbookGuiGlobal,treeView.maxJobs,self.pEmitter)
        configButton   = PConfigButton(gladefile)
        pandaMonButton = PWebButton('http://panda.cern.ch')
        savannahButton = PWebButton('https://savannah.cern.ch/projects/panda/')
        updateButton   = PUpdateButton(pbookCore,pbookGuiGlobal,self.pEmitter)
        killButton     = PKillButton(pbookCore,pbookGuiGlobal,self.pEmitter)
        retryButton    = PRetryButton(pbookCore,pbookGuiGlobal,self.pEmitter)                
        # set icons
        iconMap = {
            'retryButton'    : 'retry.png',
            'updateButton'   : 'update.png',
            'killButton'     : 'kill.png',
            'pandaMonButton' : 'pandamon.png',
            'savannahButton' : 'savannah.png',
            'configButton'   : 'config.png',
            'backButton'     : 'back.png',
            'syncButton'  : 'sync.png',
            'forwardButton'  : 'forward.png',
            }
        for buttonName in iconMap:
            iconName = iconMap[buttonName]
            image = gtk.Image()
            image.set_from_file(os.environ['PANDA_SYS'] + "/etc/panda/icons/"
                                + iconName)
            image.show()
            button = mainWindow.get_widget(buttonName)
            button.set_icon_widget(image)
        # set gtk_main_quit handler
        signal_dic = { "gtk_main_quit"             : overallQuit,
                       "on_syncButton_clicked"     : syncButton.on_clicked,
                       "on_forwardButton_clicked"  : forwardButton.on_clicked,
                       "on_backButton_clicked"     : backButton.on_clicked,
                       "on_configButton_clicked"   : configButton.on_clicked,
                       "on_pandaMonButton_clicked" : pandaMonButton.on_clicked,
                       "on_savannahButton_clicked" : savannahButton.on_clicked,
                       "on_updateButton_clicked"   : updateButton.on_clicked,
                       "on_killButton_clicked"     : killButton.on_clicked,
                       "on_retryButton_clicked"    : retryButton.on_clicked,                       
                       }
        mainWindow.signal_autoconnect(signal_dic)
        # set logger
        PLogger.setLogger(statView)
        # connect signales
        self.pEmitter.connect('on_triggerSync',  syncButton.on_clicked)
        self.pEmitter.connect('on_syncEnd',      treeView.reloadJobList)
        self.pEmitter.connect('on_rangeChanged', treeView.reloadJobList)
        self.pEmitter.connect('on_setNewJob',    sumView.showJobInfo)
        # set timer for initial sync
        gtk.timeout_add(1000, self.runSynchronizer)

        
    # run synchronizer in background
    def runSynchronizer(self):
        self.pEmitter.emit("on_triggerSync")
        return False



# extraction from myproxyUtils

import re
import sys
import getpass
from . import Client
from . import MiscUtils
from .MiscUtils import commands_get_status_output, commands_get_output


"""Classes with methods to interact with a myproxy server. 

   * Class MyProxyInterface attributes:
       - servername      = name of the myproxy server
       - vomsattributes  = voms attributes, needed to use glexec
       - userDN          = DN of the user
       - pilotownerDN    = DN of the pilot owner 
                           who will be allowed to retrieve the proxy
       - proxypath       = location for the retrieved proxy
       - pilotproxypath  = location of the pilot owner proxy

       - command         = command, for testing purposes      

       - Two dates, in seconds since Epoch, to compare with
         the version of the  myproxy tools installed on the WN 

          - myproxyinit_refdate  = 1181620800 = 12 Jun 2007
          - myproxylogon_refdate = 1181620800 = 12 Jun 2007 

       - automatic_renewal = True/False to determine if the proxy 
                             has to be periodically retrieved
       - time_threshold    = time threshold to trigger a new retrieval
                              

   * Class MyProxyInterface methods:
      - delegate: to delegate a valid proxy using command myproxy-init 
      - retrieve: to recover the stored proxy using command myproxy-logon
      - delproxy: to delete the retrieved proxy

   Proxies must to have voms attributes, to be used by glexec.
   
   user's DN is used as username. This DN can be got
   from the job's schema, defined in JobSpec.py file,
   and passed by a dictionary variable named param[]

   Proxies are stored without password.

   Only the pilot onwer is allowed to retrieve the proxy.
   His DN is specified when the proxy is uploaded, 
   using the -Z option.

   
   * Class MyProxyError is implemented to throw exceptions
   when something is going wrong.      


   * Functions to manage the proxy retrieval and the gLExec preparation
     have been moved from SimplePilot.py to here for clarity
   

   Author:         Jose Caballero (Brookhaven National Laboratory)
   E-mail:         jcaballero (at) bnl.gov
   Last revision:  5/22/2008 
   Version:        1.5
"""

class MyProxyError(Exception):
    """class to throw exceptions when something is going wrong
    during the delegation/retrieval of the proxy credentials
    to/from a myproxy server
    """

    #constructor
    def __init__(self, index, message=''):
        """index is the key of a dictionary 
        with strings explaining what happened

        message is the output returned from the application which failed        
        """
        self.__index = index      
        self.__message = message

        self.__err={}

        self.__err[2100] = ' server name not specified\n'
        self.__err[2101] = ' voms attributes not specified\n'
        self.__err[2102] = ' user DN not specified\n'
        self.__err[2103] = ' pilot owner DN not specified\n'
        self.__err[2104] = ' invalid path for the delegated proxy\n'
        self.__err[2105] = ' invalid pilot proxy path\n'
        self.__err[2106] = ' no path to delegated proxy specified\n'

        self.__err[2200] = ' myproxy-init not available in PATH\n'
        self.__err[2201] = ' myproxy-logon not available in PATH\n'

        self.__err[2202] = ' myproxy-init version not valid\n'
        self.__err[2203] = ' myproxy-logon version not valid\n'

        self.__err[2300] = ' proxy delegation failed\n'
        self.__err[2301] = ' proxy retrieval failed\n'

        self.__err[2400] = ' security violation. Logname and DN do not match\n'

        # output message
        self.__output = '\n>>> MyProxyError %s:' % self.__index + self.__err[self.__index]

        if self.__message != '':
           self.__output += '\n>>> MyProxyError: Begin of error message\n'
           self.__output += self.__message
           self.__output += '\n>>> MyProxyError: End of error message\n'

    def __str__(self):
        return self.__output

    # method to read the index
    def getIndex(self):
        return self.__index
    index = property(getIndex)


class MyProxyInterface(object):
    """class with basic features to delegate/retrieve 
    a valid proxy credential to/from a myproxy server
    """

    # constructor
    def __init__(self):
        self.__servername = ''      # name of the myproxy server
        self.__vomsattributes = 'atlas'  # voms attributes, needed to use glexec
        self.__userDN = ''          # DN of the user
        self.__pilotownerDN = ''    # DN of the pilot owner 
                                    # who will be allowed to retrieve the proxy
        self.__proxypath = ''       # location for the retrieved proxy
        self.__pilotproxypath = ''  # location of the pilot owner proxy
        
        self.__command = ''         # command, for testing purposes

        # dates, in seconds since Epoch, used as reference for myproxy tools
        # to compare with version installed on the WN
        # Reference value is 12 Jun 2007
        # which corresponds with 1181620800 seconds
        self.__myproxyinit_refdate  = 1181620800
        self.__myproxylogon_refdate = 1181620800

        # variables to manage the periodically retrieval
        # by a separate script
        self.__automatic_retrieval = 0  # by default, the option is disabled. 
        self.__time_threshold = 3600    # by default, 1 hour
        # src file to setup grid runtime
        self.srcFile = Client._getGridSrc()

    # methods to write/read the name of the myproxy server
    def setServerName(self,value):
        self.__servername = value
    def getServerName(self):
        return self.__servername
    servername = property( getServerName, setServerName )
    
    # methods to write/read the voms attributes 
    def setVomsAttributes(self,value):
        self.__vomsattributes = value
    def getVomsAttributes(self):
        return self.__vomsattributes
    vomsattributes = property( getVomsAttributes, setVomsAttributes )

    # methods to write/read the DN of the user
    # This value can be got from dictionary param
    # implementing the job schema (file JobSpec.py) 

    def __processDN_xtrastrings(self,value):
        # delegated proxies do not include strings like "/CN=proxy" or 
        # "/CN=12345678" in the DN. Extra strings have to be removed
        if value.count('/CN=') > 1:
            first_index = value.find('/CN=')
            second_index = value.find('/CN=', first_index+1)
            value = value[0:second_index]

        return value

    def __processDN_whitespaces(self,value):
        # usually the last part of the DN is like 
        #   '/CN=<Name> <LastName> <RandNumber>'
        # with white spaces. It must be 
        #   '/CN=<Name>\ <LastName>\ <RandNumber>'
        # to scape the white spaces
        pattern = re.compile('[^\\\]\s') # a whitespace preceded 
                                         # by anything but \
        if pattern.search(value):
            value = value.replace(' ', '\ ')

        return value

    def __processDN_parenthesis(self,value):

        # the DN could contain parenthesis characteres
        # They have to be preceded by a backslash also
        pattern = re.compile( '[^\\\]\(' ) # a parenthesis "(" preceded
                                           # by anything but \
        if pattern.search(value):
            value = value.replace( '(', '\(' )

        pattern = re.compile( '[^\\\]\)' ) # a parenthesis ")" preceded
                                           # by anything but \
        if pattern.search(value):
            value = value.replace( ')', '\)' )

        return value

    def __processDN(self,value):
      
        value = self.__processDN_xtrastrings(value)
        value = self.__processDN_whitespaces(value)
        value = self.__processDN_parenthesis(value)
        
        return value


    def setUserDN(self,value):
        self.__userDN = self.__processDN(value) 
    def getUserDN(self):
        return self.__userDN
    userDN = property( getUserDN, setUserDN )

    # methods to write/read the DN of the pilot owner 
    def setPilotOwnerDN(self,value):
        self.__pilotownerDN = value
    def getPilotOwnerDN(self):
        return self.__pilotownerDN
    pilotownerDN = property( getPilotOwnerDN, setPilotOwnerDN )

    # methods to write/read the location of the retrieved proxy 
    def setProxyPath(self,value):
        # checking the path is valid. 
        # It has to be a file in the /tmp directory:
        if value.startswith('/tmp'):
            self.__proxypath = value
        else:
            raise MyProxyError(2104)

    def getProxyPath(self):
        # A file in the /tmp directory is created if no
        # other path has been set already  
        if self.__proxypath == '':
            self.__proxypath = commands_get_output( 'mktemp' )

        return self.__proxypath

    proxypath = property( getProxyPath, setProxyPath )

    # methods to write/read the path to the pilot owner proxy 
    def setPilotProxyPath(self,value):
        self.__pilotproxypath = value

    def getPilotProxyPath(self):
        # checking the path is valid
        st,out = commands_get_status_output('grid-proxy-info -exists -file %s'
                                                   % self.__pilotproxypath)
        if st != 0: #invalid path
            print("\nError: not valid proxy in path %" % self.__pilotproxypath)
            print("\nUsing the already existing proxy to get the path")
            st, path = commands_get_status_output('grid-proxy-info -path')
            if st == 0:
                self.__pilotproxypath = path
            else:
                raise MyProxyError(2105)

        if self.__pilotproxypath == '':
            print("\nWarning: not valid pilot proxy path specified already")
            print("\nUsing the already existing proxy to get the path")
            st, path = commands_get_status_output('grid-proxy-info -path')
            if st == 0:
                self.__pilotproxypath = path
            else:
                raise MyProxyError(2105)

        return self.__pilotproxypath

    pilotproxypath = property( getPilotProxyPath, setPilotProxyPath )

    # method to read the command (read-only variable)
    # for testing purposes
    def getCommand(self):
         return self.__command
    command = property( getCommand ) 

    # methods to read the reference dates 
    # (read-only variables)
    def getMyproxyinitRefdate(self):
         return self.__myproxyinit_refdate
    myproxyinit_refdate = property( getMyproxyinitRefdate )

    def getMyproxylogonRefdate(self):
         return self.__myproxylogon_refdate
    myproxylogon_refdate = property( getMyproxylogonRefdate )
 
    # methods to write/read the value of automatic_retrieval variable 
    def setAutomaticRetrieval(self,value):
        if value==0 or value==1:
          self.__automatic_retrieval = value
    def getAutomaticRetrieval(self):
        return self.__automatic_retrieval
    automatic_retrieval = property( getAutomaticRetrieval, setAutomaticRetrieval )
    

    # methods to write/read the number of seconds used as time threshold 
    # to the automatic retrieval 
    def setTimeThreshold(self,value):
        self.__time_threshold = value
    def getTimeThreshold(self):
        return self.__time_threshold
    time_threshold = property( getTimeThreshold, setTimeThreshold )
    

    def delegate(self,gridPassPhrase='',verbose=False):
        """method to upload a valid proxy credential in a myproxy server.
           The proxy must to have voms attributes.
           Only the pilot owner will be allowed to retrieve the proxy,
           specified by -Z option.
           The DN of the user is used as username.
        """

        if self.servername == '' :
            raise MyProxyError(2100)
        if self.vomsattributes == '' :
            raise MyProxyError(2101)
        if self.pilotownerDN == '' :
            raise MyProxyError(2103)    

        cmd  = 'myproxy-init'

        # credname
        credname = re.sub('-','',MiscUtils.wrappedUuidGen())

        print("=== upload proxy for glexec")
        # command options
        cmd += ' -s %s'       % self.servername     # myproxy sever name
        cmd += " -x -Z '%s'"  % self.pilotownerDN   # only user with this DN 
                                                    # is allowed to retrieve it
        cmd += ' --voms %s'   % self.vomsattributes # voms attributes
        cmd += ' -d'                                # uses DN as username
        cmd += ' --credname %s' % credname
        if gridPassPhrase == '':
            if sys.stdin.isatty():  
                gridPassPhrase = getpass.getpass('Enter GRID pass phrase:')
            else:
                sys.stdout.write('Enter GRID pass phrase:')
                sys.stdout.flush()
                gridPassPhrase = sys.stdin.readline().rstrip()
                print('')
            gridPassPhrase = gridPassPhrase.replace('$','\$')
        cmd = 'echo "%s" | %s -S' % (gridPassPhrase,cmd)
        cmd = self.srcFile + ' unset GT_PROXY_MODE; ' + cmd   
            
        self.__command = cmd  # for testing purpose

        if verbose:
            cmdStr = cmd
            if gridPassPhrase != '':
                cmdStr = re.sub(gridPassPhrase,'*****',cmd)
            print(cmdStr)
        status,out = commands_get_status_output( cmd )
        if verbose:
            print(out)
        if status != 0:
            if out.find('Warning: your certificate and proxy will expire') == -1:
                if not verbose:
                    print(out)
                raise MyProxyError(2300)
        return credname


    # check proxy
    def check(self,credname,verbose=False):
        # construct command to get MyProxy credentials 
        cmd = self.srcFile + ' myproxy-info -d '
        cmd += '-s %s' % self.servername
        if verbose:
            print(cmd)
        status,output = commands_get_status_output(cmd)
        if verbose:
            print(output)
        # check timeleft
        credSector = False
        for line in output.split('\n'):
            # look for name:
            match = re.search('^\s+name:\s+([a-zA-Z0-9]+)',line)
            if match is not None:
                if match.group(1) == credname:
                    credSector = True
                else:
                    credSector = False
                continue
            if not credSector:
                continue
            # look for timeleft:
            match = re.search('^\s+timeleft:\s+([0-9:]+)',line)
            if match is not None:
                hour = match.group(1).split(':')[0]
                hour = int(hour)
                # valid more than 3 days
                if hour > 24*3:
                    return True
                return False
        # not found
        return False



from . import Client


# API call class
class IddsApi(object):

    def __init__(self, name, dumper, verbose, idds_host, compress, manager, loader):
        self.name = name
        if idds_host is not None:
            self.name += '+{}'.format(idds_host)
        self.dumper = dumper
        self.verbose = verbose
        self.compress = compress
        self.manager = manager
        self.loader = loader

    def __call__(self, *args, **kwargs):
        return Client.call_idds_command(self.name, args, kwargs, self.dumper, self.verbose, self.compress,
                                        self.manager, self.loader)


# interface to API
class IddsApiInteface(object):
    def __init__(self):
        self.dumper = None
        self.loader = None

    def __getattr__(self, item):
        return IddsApi(item, self.dumper, self.verbose, self.idds_host, self.compress, self.manager,
                       self.loader)

    def setup(self, dumper, verbose, idds_host, compress, manager, loader):
        self.dumper = dumper
        self.verbose = verbose
        self.idds_host = idds_host
        self.compress = compress
        self.manager = manager
        self.loader = loader


# entry for API
api = IddsApiInteface()
del IddsApiInteface


def get_api(dumper=None, verbose=False, idds_host=None, compress=True, manager=False, loader=None):
    """Get an API object to access iDDS through PanDA

       args:
           dumper: function object to dump json-serialized data
           verbose: True to see verbose messages
           idds_host: iDDS hostname
           compress: True to compress request body
           manager: True to use ClientManager API
           loader: function object to load json-serialized data
       return:
           an API object
    """
    api.setup(dumper, verbose, idds_host, compress, manager, loader)
    return api



"""
local job specification

"""

import re
try:
    from urllib import quote, unquote
except ImportError:
    from urllib.parse import quote, unquote
import datetime
try:
    long()
except Exception:
    long = int
try:
    unicode()
except Exception:
    unicode = str


class LocalJobSpec(object):
    # attributes
    _attributes = ('id','JobID','PandaID','jobStatus','site','cloud','jobType',
                   'jobName','inDS','outDS','libDS','provenanceID','creationTime',
                   'lastUpdate','jobParams','dbStatus','buildStatus','retryID',
                   'commandToPilot')
    # appended attributes
    appended = {
        'groupID'       : 'INTEGER',
        'releaseVar'    : 'VARCHAR(128)',
        'cacheVar'      : 'VARCHAR(128)',
        'retryJobsetID' : 'INTEGER',
        'parentJobsetID': 'INTEGER',
        'mergeJobStatus': 'VARCHAR(20)',
        'mergeJobID'    : 'TEXT',
        'nRebro'        : 'INTEGER',
        'jediTaskID'    : 'INTEGER',
        'taskStatus'    : 'VARCHAR(16)',
        }

    _attributes += tuple(appended.keys())
    # slots
    __slots__ = _attributes + ('flag_showSubstatus','flag_longFormat')


    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            setattr(self,attr,None)
        self.flag_showSubstatus = ''
        self.flag_longFormat = False

    # string format
    def __str__(self):
        # job status
        statusMap = {}
        for item in self.jobStatus.split(','):
            match = re.search('^(\w+)\*(\d+)$',item)
            if match is None:
                # non compact
                if item not in statusMap:
                    statusMap[item] = 0
                statusMap[item] += 1
            else:
                # compact
                tmpStatus = match.group(1)
                tmpCount  = int(match.group(2))
                if tmpStatus not in statusMap:
                    statusMap[tmpStatus] = 0
                statusMap[tmpStatus] += tmpCount
        # show PandaIDs in particular states
        pandaIDstatusMap = {}
        if self.flag_showSubstatus != '':
            # get PandaIDs for each status
            tmpStatusList  = self.jobStatus.split(',')
            tmpPandaIDList = self.PandaID.split(',')
            for tmpIndex,tmpPandaID in enumerate(tmpPandaIDList):
                if tmpIndex < len(tmpStatusList):
                    tmpStatus = tmpStatusList[tmpIndex]
                else:
                    # use unkown for out-range
                    tmpStatus = 'unknown'
                # status of interest
                if tmpStatus not in self.flag_showSubstatus.split(','):
                    continue
                # append
                if tmpStatus not in pandaIDstatusMap:
                    pandaIDstatusMap[tmpStatus] = 'PandaID='
                pandaIDstatusMap[tmpStatus] += '%s,' % tmpPandaID
        statusStr = self.dbStatus
        for tmpStatus in statusMap:
            tmpCount = statusMap[tmpStatus]
            statusStr += '\n%8s   %10s : %s' % ('',tmpStatus,tmpCount)
            if self.flag_showSubstatus:
                if tmpStatus in pandaIDstatusMap:
                    statusStr += '\n%8s   %10s   %s' % ('','',pandaIDstatusMap[tmpStatus][:-1])
        # disable showSubstatus
        self.flag_showSubstatus = ''
        # number of jobs
        nJobs = len(self.PandaID.split(','))
        if self.buildStatus != '':
            # including buildJob
            nJobsStr = "%d + 1(build)" % (nJobs-1)
        else:
            nJobsStr = "%d" % nJobs
        # remove duplication in inDS and outDS
        strInDS = ''
        try:
            tmpInDSList = []
            for tmpItem in str(self.inDS).split(','):
                if tmpItem not in tmpInDSList:
                    tmpInDSList.append(tmpItem)
                    strInDS += '%s,' % tmpItem
            strInDS = strInDS[:-1]
        except Exception:
            pass
        strOutDS = ''
        try:
            tmpOutDSList = []
            for tmpItem in str(self.outDS).split(','):
                if tmpItem not in tmpOutDSList:
                    tmpOutDSList.append(tmpItem)
                    strOutDS += '%s,' % tmpItem
            strOutDS = strOutDS[:-1]
        except Exception:
            pass
        # parse
        relStr = ''
        if self.releaseVar not in ['','NULL','None',None]:
            relStr = self.releaseVar
        # cache
        cacheStr = ''
        if self.cacheVar not in ['','NULL','None',None]:
            cacheStr = self.cacheVar
        # string representation
        strFormat = "%15s : %s\n"
        strOut =  ""
        strOut += strFormat % ("JobID",        self.JobID)
        if self.groupID in ['','NULL',0,'0',-1,'-1']:
            strOut += strFormat % ("JobsetID",     '')
        else:
            strOut += strFormat % ("JobsetID", self.groupID)
        strOut += strFormat % ("type",         self.jobType)
        strOut += strFormat % ("release",      relStr)
        strOut += strFormat % ("cache",        cacheStr)
        strOut += strFormat % ("PandaID",      self.encodeCompact()['PandaID'])
        strOut += strFormat % ("nJobs",        nJobsStr)
        strOut += strFormat % ("site",         self.site)
        strOut += strFormat % ("cloud",        self.cloud)
        strOut += strFormat % ("inDS",         strInDS)
        strOut += strFormat % ("outDS",        strOutDS)
        strOut += strFormat % ("libDS",        str(self.libDS))
        strOut += strFormat % ("retryID",      self.retryID)
        strOut += strFormat % ("provenanceID", self.provenanceID)
        if self.mergeJobStatus not in ['NA']:
            strOut += strFormat % ("mergeJobStatus", self.mergeJobStatus)
            strOut += strFormat % ("mergeJobID",     self.mergeJobID)
        strOut += strFormat % ("creationTime", self.creationTime.strftime('%Y-%m-%d %H:%M:%S'))
        strOut += strFormat % ("lastUpdate",   self.lastUpdate.strftime('%Y-%m-%d %H:%M:%S'))
        strOut += strFormat % ("params",       self.jobParams)
        strOut += strFormat % ("jobStatus",    statusStr)
        # return
        return strOut


    # override __getattribute__ for SQL
    def __getattribute__(self,name):
        ret = object.__getattribute__(self,name)
        if ret is None:
            return "NULL"
        return ret


    # pack tuple into JobSpec
    def pack(self,values):
        for i in range(len(self._attributes)):
            attr= self._attributes[i]
            val = values[i]
            setattr(self,attr,val)
        # expand compact values
        self.decodeCompact()


    # return a tuple of values
    def values(self,forUpdate=False):
        # make compact values
        encVal = self.encodeCompact()
        if forUpdate:
            # for UPDATE
            retS = ""
        else:
            # for INSERT
            retS = "("
        # loop over all attributes
        for attr in self._attributes:
            if attr in encVal:
                val = encVal[attr]
            else:
                val = getattr(self,attr)
            # convert datetime to str
            if type(val) == datetime.datetime:
                val = val.strftime('%Y-%m-%d %H:%M:%S')
            # add colum name for UPDATE
            if forUpdate:
                if attr == 'id':
                    continue
                retS += '%s=' % attr
            # value
            if val == 'NULL':
                retS += 'NULL,'
            else:
                retS += "'%s'," % str(val)
        retS  = retS[:-1]
        if not forUpdate:
            retS += ')'
        return retS


    # expand compact values
    def decodeCompact(self):
        # PandaID
        pStr = ''
        for item in self.PandaID.split(','):
            match = re.search('^(\d+)-(\d+)$',item)
            if match is None:
                # non compact
                pStr += (item+',')
            else:
                # compact
                sID = long(match.group(1))
                eID = long(match.group(2))
                for tmpID in range(sID,eID+1):
                    pStr += "%s," % tmpID
        self.PandaID = pStr[:-1]
        # status
        sStr = ''
        for item in self.jobStatus.split(','):
            match = re.search('^(\w+)\*(\d+)$',item)
            if match is None:
                # non compact
                sStr += (item+',')
            else:
                # compact
                tmpStatus = match.group(1)
                tmpCount  = int(match.group(2))
                for tmpN in range(tmpCount):
                    sStr += "%s," % tmpStatus
        self.jobStatus = sStr[:-1]
        # job parameters
        self.jobParams = unquote(self.jobParams)
        # datetime
        for attr in self._attributes:
            val = getattr(self, attr)
            if not isinstance(val, (str, unicode)):
                continue
            # convert str to datetime
            match = re.search('^(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)$',val)
            if match is not None:
                tmpDate = datetime.datetime(year   = int(match.group(1)),
                                            month  = int(match.group(2)),
                                            day    = int(match.group(3)),
                                            hour   = int(match.group(4)),
                                            minute = int(match.group(5)),
                                            second = int(match.group(6)))
                setattr(self,attr,tmpDate)
        # jobsetID
        if self.groupID in ['','NULL']:
            self.groupID = 0


    # make compact values
    def encodeCompact(self,includeMerge=False):
        ret = {}
        if self.isJEDI():
            if self.taskStatus in ['finished','failed','done','broken','aborted']:
                self.dbStatus = 'frozen'
            else:
                self.dbStatus = 'running'
        # job parameters
        ret['jobParams'] = quote(self.jobParams)
        # PandaID
        pStr = ''
        sID = None
        eID = None
        tmpPandaIDs = self.PandaID.split(',')
        if includeMerge:
            tmpPandaIDs += self.mergeJobID.split(',')
        for item in tmpPandaIDs:
            if item in ['','None']:
                continue
            # convert to long
            try:
                tmpID = long(item)
            except Exception:
                sID = item
                eID = item
                break
            # set start/end ID
            if sID is None:
                sID = tmpID
                eID = tmpID
                continue
            # successive number
            if eID+1 == tmpID:
                eID = tmpID
                continue
            # jump
            if sID == eID:
                pStr += '%s,' % sID
            else:
                pStr += '%s-%s,' % (sID,eID)
            # reset
            sID = tmpID
            eID = tmpID
        # last bunch
        if sID == eID:
            pStr += '%s,' % sID
        else:
            pStr += '%s-%s,' % (sID,eID)
        ret['PandaID'] = pStr[:-1]
        if self.isJEDI():
            return ret
        # job status
        sStr = ''
        sStatus = None
        nStatus = 0
        toBeFrozen = True
        for tmpStatus in self.jobStatus.split(','):
            # check if is should be frozen
            if toBeFrozen and tmpStatus not in ['finished','failed','partial','cancelled']:
                toBeFrozen = False
            # set start status
            if sStatus is None:
                sStatus = tmpStatus
                nStatus += 1
                continue
            # same status
            if sStatus == tmpStatus:
                nStatus += 1
                continue
            # jump
            if nStatus == 1:
                sStr += '%s,' % sStatus
            else:
                sStr += '%s*%s,' % (sStatus,nStatus)
            # reset
            sStatus = tmpStatus
            nStatus = 1
        # last bunch
        if nStatus == 1:
            sStr += '%s,' % sStatus
        else:
            sStr += '%s*%s,' % (sStatus,nStatus)
        ret['jobStatus'] = sStr[:-1]
        # set merge job status
        if '--mergeOutput' in self.jobParams and self.jobType not in ['usermerge']:
            if self.mergeJobStatus not in ['NA','standby','generating','generated','aborted']:
                self.mergeJobStatus = 'standby'
        else:
            self.mergeJobStatus = 'NA'
        # set dbStatus
        if toBeFrozen:
            if self.mergeJobStatus in ['standby','generating']:
                # intermediate state while generating merge jobs
                self.dbStatus = 'running'
            else:
                self.dbStatus = 'frozen'
        else:
            if self.commandToPilot=='tobekilled':
                self.dbStatus = 'killing'
            else:
                self.dbStatus = 'running'
        # return
        return ret


    # merge job generation is active
    def activeMergeGen(self):
        if '--mergeOutput' in self.jobParams and self.mergeJobStatus in ['standby','generating'] \
               and self.jobType not in ['usermerge']:
            return True
        return False


    # return column names for INSERT or full SELECT
    def columnNames(cls):
        ret = ""
        for attr in cls._attributes:
            if ret != "":
                ret += ','
            ret += attr
        return ret
    columnNames = classmethod(columnNames)


    # check if JEDI
    def isJEDI(self):
        if self.jediTaskID in [-1,'-1','']:
            return False
        return True



import shutil
import sys
import re
import os
import shlex
import atexit

from pandaclient.Group_argparse import GroupArgParser
from pandaclient import PLogger
from pandaclient import PandaToolsPkgInfo
from pandaclient import MiscUtils
from pandaclient import Client
from pandaclient import PsubUtils
from pandaclient import PrunScript

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

try:
    unicode
except Exception:
    unicode = str


# main
def main():
    # tweak sys.argv
    sys.argv.pop(0)
    sys.argv.insert(0, 'pchain')

    usage = """pchain [options]
    """

    optP = GroupArgParser(usage=usage, conflict_handler="resolve")

    group_output = optP.add_group('output', 'output dataset/files')
    group_config = optP.add_group('config', 'single configuration file to set multiple options')
    group_submit = optP.add_group('submit', 'job submission/site/retry')
    group_expert = optP.add_group('expert', 'for experts/developers only')
    group_build = optP.add_group('build', 'build/compile the package and env setup')
    group_check = optP.add_group('check', 'check workflow description')


    optP.add_helpGroup()

    group_config.add_argument('--version', action='store_const', const=True, dest='version', default=False,
                              help='Displays version')
    group_config.add_argument('-v', action='store_const', const=True, dest='verbose', default=False,
                              help='Verbose')
    group_check.add_argument('--check', action='store_const', const=True, dest='checkOnly', default=False,
                              help='Check workflow description locally')
    group_check.add_argument('--debug', action='store_const', const=True, dest='debugCheck', default=False,
                              help='verbose mode when checking workflow description locally')

    group_output.add_argument('--cwl', action='store', dest='cwl', default=None,
                              help='Name of the main CWL file to describe the workflow')
    group_output.add_argument('--yaml', action='store', dest='yaml', default=None,
                              help='Name of the yaml file for workflow parameters')

    group_build.add_argument('--useAthenaPackages', action='store_const', const=True, dest='useAthenaPackages',
                             default=False,
                             help='One or more tasks in the workflow uses locally-built Athena packages')
    group_build.add_argument('--vo', action='store', dest='vo',  default=None,
                             help="virtual organization name")
    group_build.add_argument('--extFile', action='store', dest='extFile', default='',
                             help='root or large files under WORKDIR are not sent to WNs by default. '
                                  'If you want to send some skipped files, specify their names, '
                                  'e.g., data.root,big.tgz,*.o')

    group_output.add_argument('--outDS', action='store', dest='outDS', default=None,
                              help='Name of the dataset for output and log files')
    group_output.add_argument('--official', action='store_const', const=True, dest='official', default=False,
                              help='Produce official dataset')

    group_submit.add_argument('--noSubmit', action='store_const', const=True, dest='noSubmit', default=False,
                              help="Dry-run")
    group_submit.add_argument("-3", action="store_true", dest="python3", default=False,
                              help="Use python3")
    group_submit.add_argument('--voms', action='store', dest='vomsRoles', default=None, type=str,
                              help="generate proxy with paticular roles. "
                                   "e.g., atlas:/atlas/ca/Role=production,atlas:/atlas/fr/Role=pilot")
    group_submit.add_argument('--noEmail', action='store_const', const=True, dest='noEmail', default=False,
                              help='Suppress email notification')
    group_submit.add_argument('--prodSourceLabel', action='store', dest='prodSourceLabel', default='',
                              help="set prodSourceLabel")
    group_submit.add_argument('--workingGroup', action='store', dest='workingGroup',  default=None,
                              help="set workingGroup")

    group_expert.add_argument('--intrSrv', action='store_const', const=True, dest='intrSrv', default=False,
                              help="Please don't use this option. Only for developers to use the intr panda server")
    group_expert.add_argument('--relayHost', action='store', dest='relayHost', default=None,
                              help="Please don't use this option. Only for developers to use the relay host")

    # get logger
    tmpLog = PLogger.getPandaLogger()

    # show version
    if '--version' in sys.argv:
        print("Version: %s" % PandaToolsPkgInfo.release_version)
        sys.exit(0)

    # parse args
    options = optP.parse_args()

    # check
    for arg_name in ['cwl', 'yaml', 'outDS']:
        if not getattr(options, arg_name):
            tmpStr = "argument --{0} is required".format(arg_name)
            tmpLog.error(tmpStr)
            sys.exit(1)

    # check grid-proxy
    PsubUtils.check_proxy(options.verbose, options.vomsRoles)

    # check output name
    nickName = PsubUtils.getNickname()
    if not PsubUtils.checkOutDsName(options.outDS, options.official, nickName,
                                    verbose=options.verbose):
        tmpStr = "invalid output dataset name: %s" % options.outDS
        tmpLog.error(tmpStr)
        sys.exit(1)

    # create tmp dir
    curDir = os.getcwd()
    tmpDir = os.path.join(curDir, MiscUtils.wrappedUuidGen())
    os.makedirs(tmpDir)

    # exit action
    def _onExit(dir, del_command):
        del_command('rm -rf %s' % dir)

    atexit.register(_onExit, tmpDir, MiscUtils.commands_get_output)

    # sandbox
    if options.verbose:
        tmpLog.debug("making sandbox")
    archiveName = 'jobO.%s.tar.gz' % MiscUtils.wrappedUuidGen()
    archiveFullName = os.path.join(tmpDir, archiveName)
    extensions = ['cwl', 'yaml', 'json']
    find_opt = ' -o '.join(['-name "*.{0}"'.format(e) for e in extensions])
    tmpOut = MiscUtils.commands_get_output('find . {0} | tar cvfz {1} --files-from - '.format(find_opt, archiveFullName))

    if options.verbose:
        print(tmpOut + '\n')
        tmpLog.debug("checking sandbox")
        tmpOut = MiscUtils.commands_get_output('tar tvfz {0}'.format(archiveFullName))
        print(tmpOut + '\n')

    if not options.noSubmit:
        tmpLog.info("uploading workflow sandbox")
        if options.vo:
            use_cache_srv = False
        else:
            use_cache_srv = True
        os.chdir(tmpDir)
        status, out = Client.putFile(archiveName, options.verbose, useCacheSrv=use_cache_srv, reuseSandbox=True)

        if out.startswith('NewFileName:'):
            # found the same input sandbox to reuse
            archiveName = out.split(':')[-1]
        elif out != 'True':
            # failed
            print(out)
            tmpLog.error("Failed with %s" % status)
            sys.exit(1)
    os.chdir(curDir)
    try:
        shutil.rmtree(tmpDir)
    except Exception:
        pass

    matchURL = re.search("(http.*://[^/]+)/", Client.baseURLCSRVSSL)
    sourceURL = matchURL.group(1)

    params = {'taskParams': {},
              'sourceURL': sourceURL,
              'sandbox': archiveName,
              'workflowSpecFile': options.cwl,
              'workflowInputFile': options.yaml,
              'language': 'cwl',
              'outDS': options.outDS,
              'base_platform': os.environ.get('ALRB_USER_PLATFORM', 'centos7')
              }

    # making task params with dummy exec
    task_type_args = {'container': '--containerImage __dummy_container__'}
    if options.useAthenaPackages:
        task_type_args['athena'] = '--useAthenaPackages'
    for task_type in task_type_args:
        os.chdir(curDir)
        prun_exec_str = '--exec __dummy_exec_str__ --outDS {0} {1}'.format(options.outDS, task_type_args[task_type])
        if options.noSubmit:
            prun_exec_str += ' --noSubmit'
        if options.verbose:
            prun_exec_str += ' -v'
        if options.vo:
            prun_exec_str += ' --vo {0}'.format(options.vo)
        if options.prodSourceLabel:
            prun_exec_str += ' --prodSourceLabel {0}'.format(options.prodSourceLabel)
        if options.workingGroup:
            prun_exec_str += ' --workingGroup {0}'.format(options.workingGroup)
        if options.extFile:
            prun_exec_str += ' --extFile {0}'.format(options.extFile)
        arg_dict = {'get_taskparams': True,
                    'ext_args': shlex.split(prun_exec_str)}
        if options.checkOnly:
            arg_dict['dry_mode'] = True

        taskParamMap = PrunScript.main(**arg_dict)
        del taskParamMap['noInput']
        del taskParamMap['nEvents']
        del taskParamMap['nEventsPerJob']

        params['taskParams'][task_type] = taskParamMap

    if options.noSubmit:
        if options.noSubmit:
            if options.verbose:
                tmpLog.debug("==== taskParams ====")
                tmpKeys = list(taskParamMap)
                tmpKeys.sort()
                for tmpKey in tmpKeys:
                    if tmpKey in ['taskParams']:
                        continue
                    print('%s : %s' % (tmpKey, taskParamMap[tmpKey]))
        sys.exit(0)

    data = {'relay_host': options.relayHost, 'verbose': options.verbose}
    if not options.checkOnly:
        action_type = 'submit'
    else:
        action_type = 'check'
        data['check'] = True

    # set to use INTR server just before taking action so that sandbox files go to the regular place
    if options.intrSrv:
        Client.useIntrServer()

    # action
    tmpLog.info("{0} workflow {1}".format(action_type, options.outDS))
    tmpStat, tmpOut = Client.send_workflow_request(params, **data)

    # result
    exitCode = None
    if tmpStat != 0:
        tmpStr = "workflow {0} failed with {1}".format(action_type, tmpStat)
        tmpLog.error(tmpStr)
        exitCode = 1
    if tmpOut[0]:
        if not options.checkOnly:
            tmpStr = tmpOut[1]
            tmpLog.info(tmpStr)
        else:
            check_stat = tmpOut[1]['status']
            check_log = 'messages from the server\n' + tmpOut[1]['log']
            tmpLog.info(check_log)
            if check_stat:
                tmpLog.info('successfully verified workflow description')
            else:
                tmpLog.error('workflow description is corrupted')
    else:
        tmpStr = "workflow {0} failed. {1}".format(action_type, tmpOut[1])
        tmpLog.error(tmpStr)
        exitCode = 1
    return exitCode



#!/usr/bin/env python
import sys
try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote
import xml.dom.minidom

class dom_job:
    """ infiles[inds]=[file1,file2...]
        outfiles = [file1,file2...]
        command  - script that will be executed on the grid
        prepend  - list of (option,value) prepended to output file name
        forward  - list of (option,value) forwarded to the grid job
    """
    def __init__(s,domjob=None,primaryds=None,defaultcmd=None,defaultout=[]):
        """ Loads <job></job> from xml file.
        If primaryds is set, makes sure it is present in job spec """
        s.infiles = {}
        s.outfiles = []
        s.command=defaultcmd
        s.prepend  = []
        s.forward = []
        if not domjob:
            return
        # script executed on the grid node for this job
        if len(domjob.getElementsByTagName('command'))>0:
            s.command = dom_parser.text(domjob.getElementsByTagName('command')[0])
        # input files
        for inds in domjob.getElementsByTagName('inds'):
            name = dom_parser.text(inds.getElementsByTagName('name')[0])
            files = inds.getElementsByTagName('file')
            if len(files)==0:
                continue
            s.infiles[name]=[]
            for file in files:
                s.infiles[name].append(dom_parser.text(file))
        if primaryds and primaryds not in s.infiles.keys():
            print('ERROR: primaryds=%s must be present in each job'%primaryds)
            sys.exit(0)
        # output files (also, drop duplicates within this job)
        outfiles = set(defaultout)
        [outfiles.add(dom_parser.text(v)) for v in domjob.getElementsByTagName('output')]
        s.outfiles = list(outfiles)
        # gearing options
        for o in domjob.getElementsByTagName('option'):
            name=o.attributes['name'].value
            value=dom_parser.text(o)
            prepend=dom_parser.true(o.attributes['prepend'].value)
            forward=dom_parser.true(o.attributes['forward'].value)
            if prepend:
                s.prepend.append((name,value))
            if forward:
                s.forward.append((name,value))
    def to_dom(s):
        """ Converts this job to a dom tree branch """
        x = xml.dom.minidom.Document()
        job = x.createElement('job')
        for inds in s.infiles.keys():
            job.appendChild(x.createElement('inds'))
            job.childNodes[-1].appendChild(x.createElement('name'))
            job.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(inds))
            for file in s.infiles[inds]:
                job.childNodes[-1].appendChild(x.createElement('file'))
                job.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(file))
        for outfile in s.outfiles:
            job.appendChild(x.createElement('output'))
            job.childNodes[-1].appendChild(x.createTextNode(outfile))
        if s.command:
            job.appendChild(x.createElement('command'))
            job.childNodes[-1].appendChild(x.createTextNode(s.command))
        for option in s.prepend + list(set(s.prepend+s.forward)-set(s.prepend)):
            job.appendChild(x.createElement('option'))
            job.childNodes[-1].setAttribute('name',str(option[0]))
            if option in s.forward:
                job.childNodes[-1].setAttribute('forward','true')
            else:
                job.childNodes[-1].setAttribute('forward','false')
            if option in s.prepend:
                job.childNodes[-1].setAttribute('prepend','true')
            else:
                job.childNodes[-1].setAttribute('prepend','false')
            job.childNodes[-1].appendChild(x.createTextNode(str(option[1])))
        return job
    def files_in_DS(s,DS):
        """ Returns a list of files used in a given job in a given dataset"""
        if DS in s.infiles:
            return s.infiles[DS]
        else:
            return []
    def forward_opts(s):
        """ passable string of forward options """
        return ' '.join( ['%s=%s'%(v[0],v[1]) for v in s.forward])
    def prepend_string(s):
        """ a tag string prepended to output files """
        return '_'.join( ['%s%s'%(v[0],v[1]) for v in s.prepend])
    def exec_string(s):
        """ exec string for prun.
        If user requested to run script run.sh (via <command>run.sh</command>), it will return
        opt1=value1 opt2=value2 opt3=value3 run.sh
        This way, all options will be set inside run.sh
        """
        return '%s %s'%(s.forward_opts(),s.command)
    def exec_string_enc(s):
        """ exec string for prun.
        If user requested to run script run.sh (via <command>run.sh</command>), it will return
        opt1=value1 opt2=value2 opt3=value3 run.sh
        This way, all options will be set inside run.sh
        """
        comStr = '%s %s'%(s.forward_opts(),s.command)
        return quote(comStr)
    def get_outmap_str(s,outMap):
        """ return mapping of original and new filenames 
        """
        newMap = {}
        for oldLFN in outMap:
            fileSpec = outMap[oldLFN]
            newMap[oldLFN] = str(fileSpec.lfn)
        return str(newMap)
    def outputs_list(s,prepend=False):
        """ python list with finalized output file names """
        if prepend and s.prepend_string():
            return [s.prepend_string()+'.'+o for o in s.outfiles]
        else:
            return [o for o in s.outfiles]
    def outputs(s,prepend=False):
        """ Comma-separated list of output files accepted by prun """
        return ','.join(s.outputs_list(prepend))

class dom_parser:
    def __init__(s,fname=None,xmlStr=None):
        """ creates a dom object out of a text file (if provided) """
        s.fname = fname
        s.dom = None
        s.title = None
        s.tag = None
        s.command = None
        s.outds = None
        s.inds = {}
        s.global_outfiles = []
        s.jobs = []
        s.primaryds = None
        if fname:
            s.dom = xml.dom.minidom.parse(fname)
            s.parse()
            s.check()
        if xmlStr is not None:
            s.dom = xml.dom.minidom.parseString(xmlStr)
            s.parse()
            s.check()

    @staticmethod
    def break_regex(v,N=100):
        """ breaks up a very long regex into a comma-separeted list of filters """
        _REGEXLIM=2**15-1000
        spl=v.split('|')
        res=[]
        for i in range(0,len(spl),N):
            piece = '|'.join(spl[i:i+N])
            assert len(piece)<_REGEXLIM,'Input dataset contains very long filenames. You must reduce parameter N in break_regex()'
            res.append( piece )
        return ','.join(res)
    @staticmethod
    def true(v):
        """ define True """
        return v in ('1', 'true', 'True', 'TRUE', 'yes', 'Yes', 'YES')
    @staticmethod
    def text(pnode):
        """ extracts the value stored in the node """
        rc = []
        for node in pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc.append(str(node.data).strip())
        return ''.join(rc)
    def parse(s):
        """ loads submission configuration from an xml file """
        try:
            # general settings
            if len(s.dom.getElementsByTagName('title'))>0:
                s.title = dom_parser.text(s.dom.getElementsByTagName('title')[0])
            else:
                s.title = 'Default title'
            if len(s.dom.getElementsByTagName('tag'))>0:
                s.tag = dom_parser.text(s.dom.getElementsByTagName('tag')[0])
            else:
                s.tag = 'default_tag'
            s.command = None  # can be overridden in subjobs
            for elm in s.dom.getElementsByTagName('submission')[0].childNodes:
                if elm.nodeName != 'command':
                    continue
                s.command = dom_parser.text(elm)
                break
            s.global_outfiles = []    # subjobs can append *additional* outputs
            for elm in s.dom.getElementsByTagName('submission')[0].childNodes:
                if elm.nodeName != 'output':
                    continue
                s.global_outfiles.append(dom_parser.text(elm))
            s.outds = dom_parser.text(s.dom.getElementsByTagName('outds')[0])
            # declaration of all input datasets
            primarydss = []
            for elm in s.dom.getElementsByTagName('submission')[0].childNodes:
                if elm.nodeName != 'inds':
                    continue
                if 'primary' in elm.attributes.keys() and dom_parser.true(elm.attributes['primary'].value):
                    primary=True
                else:
                    primary=False
                stream=dom_parser.text(elm.getElementsByTagName('stream')[0])
                name=dom_parser.text(elm.getElementsByTagName('name')[0])
                s.inds[name]=stream
                if primary:
                    primarydss.append(name)
            # see if one of the input datasets was explicitly labeled as inDS
            if len(primarydss)==1:
                s.primaryds = primarydss[0]
            else:
                s.primaryds = None
            for job in s.dom.getElementsByTagName('job'):
                s.jobs.append(dom_job(job,primaryds=s.primaryds,defaultcmd=s.command,defaultout=s.global_outfiles))
        except Exception:
            print('ERROR: failed to parse',s.fname)
            raise
    def to_dom(s):
        """ Converts this submission to a dom tree branch """
        x = xml.dom.minidom.Document()
        submission = x.createElement('submission')
        if s.title:
            submission.appendChild(x.createElement('title'))
            submission.childNodes[-1].appendChild(x.createTextNode(s.title))
        if s.tag:
            submission.appendChild(x.createElement('tag'))
            submission.childNodes[-1].appendChild(x.createTextNode(s.tag))
        for name in s.inds:
            stream = s.inds[name]
            submission.appendChild(x.createElement('inds'))
            if name==s.primaryds:
                submission.childNodes[-1].setAttribute('primary','true')
            else:
                submission.childNodes[-1].setAttribute('primary','false')
            submission.childNodes[-1].appendChild(x.createElement('stream'))
            submission.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(stream))
            submission.childNodes[-1].appendChild(x.createElement('name'))
            submission.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(name))
        if s.command:
            submission.appendChild(x.createElement('command'))
            submission.childNodes[-1].appendChild(x.createTextNode(s.command))
        for outfile in s.global_outfiles:
            submission.appendChild(x.createElement('output'))
            submission.childNodes[-1].appendChild(x.createTextNode(outfile))
        submission.appendChild(x.createElement('outds'))
        submission.childNodes[-1].appendChild(x.createTextNode(s.outds))
        for job in s.jobs:
            submission.appendChild(job.to_dom())
        return submission
    def check(s):
        """ checks that all output files have unique qualifiers """
        quals=[]
        for j in s.jobs:
            quals+=j.outputs_list(True)
        if len(list(set(quals))) != len(quals):
            print('ERROR: found non-unique output file names across the jobs')
            print('(you likely need to review xml options with prepend=true)')
            sys.exit(0)
    def input_datasets(s):
        """ returns a list of all used input datasets """
        DSs = set()
        for j in s.jobs:
            for ds in j.infiles.keys():
                DSs.add(ds)
        return list(DSs)
    def outDS(s):
        """ output dataset """
        return s.outds
    def inDS(s):
        """ chooses a dataset we'll call inDS; others will become secondaryDS """
        # user manually labeled one of datasets as primary, so make it inDS:
        if s.primaryds:
            return s.primaryds
        # OR: choose inDS dataset randomly
        else:
            return s.input_datasets()[0]
    def secondaryDSs(s):
        """ returns all secondaryDSs. This excludes inDS, unless inDS is managed by prun"""
        return [d for d in s.input_datasets() if d!=s.inDS() ]
    def secondaryDSs_config(s,filter=True):
        """ returns secondaryDSs string in prun format:
        StreamName:nFilesPerJob:DatasetName[:MatchingPattern[:nSkipFiles]]
        nFilesPerJob is set to zero, so that it is updated later from actual file count.
        MatchingPattern is an OR-separated list of actual file names.
        """
        out = []
        DSs = s.secondaryDSs()
        for i,DS in enumerate(DSs):
            if DS in s.inds:
                stream=s.inds[DS]
            else:
                stream='IN%d'%(i+1,)
            # remove scope: since it conflicts with delimiter (:)
            DS = DS.split(':')[-1]
            if filter:
                out.append('%s:0:%s:%s'%(stream,DS,s.files_in_DS(DS,regex=True)))
            else:
                out.append('%s:0:%s'%(stream,DS))
        return ','.join(out)
    def writeInputToTxt(s):
        """ Prepares prun option --writeInputToTxt
        comma-separated list of STREAM:STREAM.files.dat
        """
        out = []
        DSs = s.secondaryDSs()
        for i,DS in enumerate(DSs):
            if DS in s.inds:
                stream=s.inds[DS]
            else:
                stream='IN%d'%(i+1,)
            out.append('%s:%s.files.dat'%(stream,stream))
        out.append('IN:IN.files.dat')
        return ','.join(out)
    def files_in_DS(s,DS,regex=False):
        """ Returns a list of all files from a given dataset
            that will be used in at least one job in this submission
            If regex==True, the list is converted to a regex string
        """
        assert DS in s.input_datasets(),'ERROR: dataset %s was not requested in the xml file'%DS
        files = []
        for j in s.jobs:
            if DS in j.infiles.keys():
                files+=j.infiles[DS]
        if regex:
            return '|'.join(sorted(list(set(files))))
        else:
            return sorted(list(set(files)))
    def nFiles_in_DS(s,DS):
        return len(s.files_in_DS(DS))
    def nJobs(s):
        return len(s.jobs)
    def dump(s,verbose=True):
        """ prints a summary of this submission """
        def P(key,value=''):
            if value=='':
                print(key)
            else:
                print((key+':').ljust(14),)
                print(value)
        P('XML FILE LOADED',s.fname)
        P('Title',s.title)
        P('Command',s.command)
        P('InDS',s.inDS())
        P('Output DS',s.outds)
        P('njobs',s.nJobs())
        if verbose:
            for i,job in enumerate(s.jobs):
                P('===============> JOB%d'%i)
                P('command',job.exec_string())
                P('outfiles',job.outputs())
                P('INPUTS:')
                j=0
                for dsname in job.infiles:
                    files = job.infiles[dsname]
                    P('  Dataset%d'%j,dsname)
                    for k,fname in enumerate(files):
                        P('     File%d'%k,fname)
                    j+=1

if __name__=="__main__":
    p = dom_parser('./job.xml')
    p.dump()



import os
import sys
import getpass
import subprocess
from IPython.core.magic import register_line_magic

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
from . import PLogger


# setup
def setup():
    tmp_log = PLogger.getPandaLogger()
    # parse config file
    conf_file = os.path.expanduser('~/.panda/panda_setup.cfg')
    if not os.path.exists(conf_file):
        tmp_log.error('panda conifg file is missing at {}'.format(conf_file))
        return False
    parser = ConfigParser.ConfigParser()
    parser.read(conf_file)
    section = parser['main']

    # variables
    panda_install_scripts = section['PANDA_INSTALL_SCRIPTS']
    panda_install_purelib = section['PANDA_INSTALL_PURELIB']
    panda_install_dir = section['PANDA_INSTALL_DIR']

    # PATH
    paths = os.environ['PATH'].split(':')
    if not panda_install_scripts in paths:
        paths.insert(0, panda_install_scripts)
        os.environ['PATH'] = ':'.join(paths)

    # PYTHONPATH
    if 'PYTHONPATH' not in os.environ:
        os.environ['PYTHONPATH'] = panda_install_purelib
    else:
        paths = os.environ['PYTHONPATH'].split(':')
        if panda_install_purelib not in paths:
            paths.insert(0, panda_install_scripts)
            os.environ['PYTHONPATH'] = ':'.join(paths)

    # env
    panda_env = {'PANDA_CONFIG_ROOT': '~/.pathena',
                 'PANDA_SYS': panda_install_dir,
                 "PANDA_PYTHONPATH": panda_install_purelib,
                 "PANDA_VERIFY_HOST": "off",
                 "PANDA_JUPYTER": "1",
                 }
    for i in ['PANDA_AUTH',
              'PANDA_AUTH_VO',
              'PANDA_URL_SSL',
              'PANDA_URL',
              'PANDAMON_URL',
              'X509_USER_PROXY',
              'PANDA_USE_NATIVE_HTTPLIB',
              'PANDA_NICKNAME',
              ]:
        try:
            panda_env[i] = section[i]
        except Exception:
            pass
    os.environ.update(panda_env)


# magic commands

GETPASS_STRINGS = ['Enter GRID pass phrase for this identity:']
RAWINPUT_STRINGS = ['>>> \n', "[y/n] \n"]


def _execute(command):
    with subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                          stdin=subprocess.PIPE, universal_newlines=True) as p:
        while True:
            nextline = p.stdout.readline()
            if nextline == '' and p.poll() is not None:
                break
            # check if uses getpass or raw_input
            is_getpass = False
            is_raw_input = False
            for one_str in GETPASS_STRINGS:
                if one_str in nextline:
                    is_getpass = True
                    break
            for one_str in RAWINPUT_STRINGS:
                if one_str == nextline:
                    is_raw_input = True
                    break
            if not is_raw_input:
                sys.stdout.write(nextline)
                sys.stdout.flush()
            # need to call getpass or input since jupyter notebook doesn't pass stdin from subprocess
            st = None
            if is_getpass:
                st = getpass.getpass()
                p.stdin.write(st)
                p.stdin.flush()
            elif is_raw_input:
                st = input('\n' + one_str.strip())
            # feed stdin
            if st is not None:
                p.stdin.write(st + '\n')
                p.stdin.flush()

        output = p.communicate()[0]
        exit_code = p.returncode
        if exit_code == 0:
            if output:
                print(output)
        return exit_code


@register_line_magic
def pathena(line):
    _execute('pathena ' + line + ' -3')
    return


@register_line_magic
def prun(line):
    _execute('prun ' + line + ' -3')
    return


@register_line_magic
def phpo(line):
    _execute('phpo ' + line + ' -3')
    return


@register_line_magic
def pbook(line):
    _execute('pbook ' + line + ' --prompt_with_newline')
    return


del pathena, prun, phpo, pbook
